"""Module for all helper methods to evaluate algorithms."""
from tpcp.validate._scorer import Aggregator, MeanAggregator, NoAgg, Scorer
from tpcp.validate._validate import cross_validate, validate

__all__ = ["Scorer", "NoAgg", "Aggregator", "MeanAggregator", "cross_validate", "validate"]



"""Score and fit results.

This is a modified version of sklearn fit and score functionality.

The original code is licenced under BSD-3: https://github.com/scikit-learn/scikit-learn
"""

from __future__ import annotations

import time
from typing import TYPE_CHECKING, Any, Optional, Union

from joblib import Memory
from typing_extensions import TypedDict

from tpcp._base import clone
from tpcp._hash import custom_hash
from tpcp._utils._general import _get_nested_paras
from tpcp.exceptions import OptimizationError, TestError

if TYPE_CHECKING:
    from tpcp._dataset import Dataset
    from tpcp._optimize import BaseOptimize
    from tpcp._pipeline import Pipeline
    from tpcp.validate import Scorer

_SCORE_TYPE = Union[dict[str, float], float]  # pylint: disable=invalid-name
_AGG_SCORE_TYPE = Union[dict[str, float], float]  # pylint: disable=invalid-name
_SINGLE_SCORE_TYPE = Union[dict[str, list[float]], Optional[list[float]]]  # pylint: disable=invalid-name


class _ScoreResults(TypedDict, total=False):
    """Type representing results of _score."""

    scores: _AGG_SCORE_TYPE
    single_scores: _SINGLE_SCORE_TYPE
    score_time: float
    data_labels: list[Union[str, tuple[str, ...]]]
    parameters: Optional[dict[str, Any]]


class _OptimizeScoreResults(TypedDict, total=False):
    """Type representing results of _score_and_optimize."""

    test_scores: _AGG_SCORE_TYPE
    test_single_scores: _SINGLE_SCORE_TYPE
    train_scores: _AGG_SCORE_TYPE
    train_single_scores: _SINGLE_SCORE_TYPE
    score_time: float
    optimize_time: float
    train_data_labels: list[Union[str, tuple[str, ...]]]
    test_data_labels: list[Union[str, tuple[str, ...]]]
    parameters: Optional[dict[str, Any]]
    optimizer: BaseOptimize


def _score(
    pipeline: Pipeline,
    dataset: Dataset,
    scorer: Scorer,
    parameters: Optional[dict[str, Any]],
    return_parameters=False,
    return_data_labels=False,
    return_times=False,
    error_info: Optional[str] = None,
) -> _ScoreResults:
    """Set parameters and return score.

    Parameters
    ----------
    pipeline
        An instance of a tpcp pipeline
    dataset
        An instance of a tpcp dataset with multiple data points.
    scorer
        A scorer that calculates a score by running the pipeline on each data point and then aggregates the results.
    parameters : dict of valid parameters for the pipeline
        The parameters that should be set for the pipeline before scoring
    return_parameters
        If the parameter value that was inputted should be added to the result dict
    return_data_labels
        If the names of the data points should be added to the result dict
    return_times
        If the time required to score the dataset should be added to the result dict

    Returns
    -------
    result : dict with the following attributes
        scores : dict of scorer name -> float
            Calculated scores
        scores_single : dict of scorer name -> np.ndarray
            Calculated scores for each individual data point
        score_time : float
            Time required to score the dataset
        data : list
            List of data point labels used
        parameters : dict or None
            The parameters that have been evaluated.

    """
    if parameters is not None:
        # clone after setting parameters in case any parameters are estimators (like pipeline steps).
        parameters = _clone_parameter_dict(parameters)

        pipeline = pipeline.set_params(**parameters)

    try:
        start_time = time.time()
        agg_scores, single_scores = scorer(pipeline, dataset)
        score_time = time.time() - start_time
    except Exception as e:  # noqa: BLE001
        raise TestError(
            f"Testing the algorithm on the dataset failed with the error above.\n{error_info or ''}\n\n"
            f"The test-set is:\n{[d.group_labels for d in dataset]}"
        ) from e

    result: _ScoreResults = {"scores": agg_scores, "single_scores": single_scores}
    if return_times:
        result["score_time"] = score_time
    if return_data_labels:
        result["data_labels"] = dataset.group_labels
    if return_parameters:
        result["parameters"] = parameters
    return result


def _optimize_and_score(
    optimizer: BaseOptimize,
    scorer: Scorer,
    train_set: Dataset,
    test_set: Dataset,
    *,
    optimize_params: Optional[dict] = None,
    hyperparameters: Optional[dict] = None,
    pure_parameters: Optional[dict] = None,
    return_train_score=False,
    return_optimizer=False,
    return_parameters=False,
    return_data_labels=False,
    return_times=False,
    memory: Optional[Memory] = None,
    error_info: Optional[str] = None,
) -> _OptimizeScoreResults:
    """Optimize and score the optimized pipeline on the train and test data, respectively.

    This method is aware of the differences between hyperparameters and normal (pure) parameters.
    This can be used to cache the results of training, as the training results should only depend on hyperparameters,
    but not on normal parameters.
    The provided `memory` instance is used to perform this caching step.

    Note, that caching in this context should only be performed in the context of a single call to e.g. GridSearchCV
    and the cache should be deleted afterwards to avoid cache leak.
    Therefore, the cachedir should ideally be set to a random tmp dir by the caller.

    Note: In the past, we only provided the index to the optimize method and not the subsets of the data directly.
    We changed that, because in cases, where the data index is non-deterministic, the order of the dataset might
    change between the point where we calculate the split (in the caller) and the point where we extract the respective
    data points (here).
    This might even happen in a different Python process (in the case of multiprocessing) where the dataset object was
    recreated.
    Hence, it might be that the dataobject recreated after pickling and unpickling is not the same as the original one.
    To avoid this entirely, we split and extract the subset right in the same process in the caller and only pass
    the subset to this method.
    Really badly written Datasets might still get past this, but let's hope this rarely happens.
    Note, that the Dataset object itself has a safeguard against non-deterministic indices, but this is not a guarantee.
    Let's better be safe and reduce the surface for bugs even further here.

    """
    if memory is None:
        memory = Memory(None)
    # clone after setting parameters in case any parameters are estimators (like pipeline steps).
    hyperparameters = _clone_parameter_dict(hyperparameters)
    pure_parameters = _clone_parameter_dict(pure_parameters)

    optimize_params_clean: dict = optimize_params or {}

    try:
        start_time = time.time()
        optimizer = _cached_optimize(
            optimizer, train_set, hyperparameters, pure_parameters, memory, optimize_params_clean
        )
        optimize_time = time.time() - start_time
    except Exception as e:  # noqa: BLE001
        raise OptimizationError(
            f"The optimization on the trainset failed with the error above.\n{error_info or ''}\n\n"
            f"This optimization used the following trainset:\n{train_set}"
        ) from e

    # Now we set the remaining paras.
    # Because, we need to set the parameters on the optimized pipeline and not the input pipeline we strip the
    # naming prefix.
    striped_paras = _get_nested_paras(pure_parameters, "pipeline")
    optimizer.optimized_pipeline_.set_params(**striped_paras)
    # We also set the parameters of the input pipeline to make it seem that all parameters were set from the
    # beginning.
    optimizer = optimizer.set_params(**pure_parameters)

    try:
        agg_scores, single_scores = scorer(optimizer.optimized_pipeline_, test_set)
        score_time = time.time() - optimize_time - start_time
    except Exception as e:  # noqa: BLE001
        raise TestError(
            f"Testing the optimized algorithm on the test-set failed with the error above.\n{error_info or ''}\n\n"
            f"The test-set is:\n{test_set}"
        ) from e

    result: _OptimizeScoreResults = {"test_scores": agg_scores, "test_single_scores": single_scores}
    if return_train_score:
        try:
            train_agg_scores, train_single_scores = scorer(optimizer.optimized_pipeline_, train_set)
        except Exception as e:  # noqa: BLE001
            raise TestError(
                "Running the optimized algorithm on the train-set to calculate the train error failed with the error "
                f"above.\n{error_info or ''}\n\n"
                f"The train-set is:\n{[d.group_labels for d in test_set]}"
            ) from e
        result["train_scores"] = train_agg_scores
        result["train_single_scores"] = train_single_scores
    if return_times:
        result["score_time"] = score_time
        result["optimize_time"] = optimize_time
    if return_data_labels:
        # Note we always return the train data attribute as it is interesting information independent of the train
        # score and has 0 runtime impact.
        result["train_data_labels"] = train_set.group_labels
        result["test_data_labels"] = test_set.group_labels
    if return_optimizer:
        # This is the actual trained optimizer. This means that `optimizer.optimized_pipeline_` contains the actual
        # instance of the trained pipeline.
        result["optimizer"] = optimizer
    if return_parameters:
        result["parameters"] = {**hyperparameters, **pure_parameters}
    return result


def _cached_optimize(
    optimizer: BaseOptimize, data: Dataset, hyperparameters: dict, pure_parameters: dict, memory: Memory, optimize_paras
):
    """Set parameters and optimize a pipeline and cache the optimization result.

    This method will cache the training as long as the hyperparameters stay the same.
    Changing the pure parameters will not invalidate the cache.

    """

    # We do not set all paras right away, we first create a cached optimize function that only has the hyper
    # parameters as input.
    # This allows to cache the train results, if the _optimize_and_score is called multiple times with the same hyper
    # parameters.
    # To be sure that nothing "bad" happens here, we also pass in the pipeline class itself to invalidate the cache,
    # in case a completely different pipeline/algorithm is optimized.
    # Ideally the `memory` object used here should only be used once.
    # E.g. for a single a GridSearchCV.
    def cachable_optimize(
        opti: type[BaseOptimize], hyperparas: dict[str, Any], data: Dataset, optimize_params: dict
    ) -> BaseOptimize:
        _ = opti
        return optimizer.set_params(**hyperparas).optimize(data, **optimize_params)

    optimize_func = memory.cache(cachable_optimize)
    # Optimization must never modify pure parameters, or we have a problem.
    # We check that by calculating the hash of all pure parameters before and after the optimization.
    opti_paras = optimizer.get_params()
    pure_para_subset = {k: opti_paras[k] for k in pure_parameters}
    pure_para_hash = custom_hash(pure_para_subset)
    pipeline_pure_para_hash = custom_hash(_get_nested_paras(pure_para_subset, "pipeline"))

    # This is the actual call to train the optimizer:
    optimizer = optimize_func(type(optimizer), hyperparameters, data, optimize_paras)

    opti_paras = optimizer.get_params()
    optimized_pipeline_paras = optimizer.optimized_pipeline_.get_params()
    # We check that the pure parameters on the optimize object haven't changed and that the pure parameters belonging
    # to the pipeline have not changed in the `optimized_pipeline`.
    # Note, that the first case will never happen with tpcp native Optimizers, but could happen for custom
    # optimizers.
    if pipeline_pure_para_hash != custom_hash(
        {k: optimized_pipeline_paras[k] for k in _get_nested_paras(pure_parameters, "pipeline")}
    ) or pure_para_hash != custom_hash({k: opti_paras[k] for k in pure_parameters}):
        raise ValueError(
            "Optimizing the pipeline modified a parameter marked as `pure`. "
            "This must not happen. "
            "Double check your optimize implementation and the list of pure parameters."
        )

    return optimizer


def _clone_parameter_dict(param_dict: Optional[dict]) -> dict:
    cloned_param_dict = {}
    if param_dict is not None:
        for k, v in param_dict.items():
            cloned_param_dict[k] = clone(v, safe=False)
    return cloned_param_dict



"""Type only classes that should be used when trying to type tpcp derived classes.

Note, this module only imports them, but they are defined close to places where they are used.
"""


from tpcp._algorithm import AlgorithmT
from tpcp._base import BaseTpcpObjectT
from tpcp._dataset import DatasetT
from tpcp._pipeline import OptimizablePipelineT, PipelineT
from tpcp.validate._scorer import (
    AggReturnType,
    MultiScoreTypeT,
    ScoreFunc,
    ScoreFuncMultiple,
    ScoreFuncSingle,
    ScoreTypeT,
    SingleScoreTypeT,
)

__all__ = [
    "BaseTpcpObjectT",
    "PipelineT",
    "OptimizablePipelineT",
    "DatasetT",
    "AlgorithmT",
    "ScoreTypeT",
    "SingleScoreTypeT",
    "MultiScoreTypeT",
    "AggReturnType",
    "ScoreFuncSingle",
    "ScoreFuncMultiple",
    "ScoreFunc",
]



"""Public base classes for all algorithms and pipelines."""

from __future__ import annotations

from typing import ClassVar, TypeVar, Union

from tpcp._base import BaseTpcpObject

AlgorithmT = TypeVar("AlgorithmT", bound="Algorithm")


class Algorithm(BaseTpcpObject):
    """Base class for all algorithms.

    All type-specific algorithm classes should inherit from this class and need to

    1. overwrite `_action_method` with the name of the actual action method of this class type
    2. implement a stub for the action method

    If you want to create an optimizable algorithm, add a `self_optimize` or (`self_optimize_with_info`) method to your
    class.
    We do not provide a separate base class for that, as we can make no assumptions about the call signature of your
    custom `self_optimize` method.
    If you need an "optimizable" version for a group of algorithms you are working with, create a custom
    `OptimizableAlgorithm` class or `OptimizableAlgorithmMixing` that is specific to your algorithm.

    Attributes
    ----------
    _action_methods
        The name(s) of the action method used by the child class

    """

    _action_methods: ClassVar[Union[tuple[str, ...], str]]



"""A custom hash function implementation that properly supports pytorch."""
import contextlib
import io
import os
import pickle
import struct
import sys
from pathlib import Path

from joblib.hashing import Hasher, NumpyHasher

Pickler = pickle._Pickler


class NoMemoizeHasher(Hasher):
    """A joblib hasher with all memoize features disabled."""

    def memoize(self, obj):  # noqa: ARG002
        # We skip memoization here, as it can cause some issues with nested objects.
        # https://github.com/joblib/joblib/issues/1283
        # In particular with the way cloning is implemented in tpcp, such bugs might occur more often than in other
        # applications.
        # My understanding is that memoization is used to reduce the size of the output.
        # Not using memoization is actually faster, but will fail with self referential objects.
        # (https://docs.python.org/3/library/pickle.html#pickle.Pickler.fast).
        # In these cases hashing will throw an `RecursionError`.
        # So it seems like it is a tradeoff between the two issues.
        # For now, we accept the recursion issue, as I think this might happen less often by accident.
        return

    def hash(self, obj, return_digest=True):  # noqa: A003
        """Get hash while handling some edgecases.

        Namely, this implementation fixes the following issues:

        - Because we skip memoization, we need to handle the case where the object is self-referential.
          We just catch the error and raise a more descriptive error message.


        """
        try:
            return super().hash(obj, return_digest)
        except RecursionError as e:
            raise ValueError(
                "The custom hasher used in tpcp does not support hashing of self-referential objects."
            ) from e

    def save_global(self, obj, name=None, pack=struct.pack):
        """We overwrite this method to fix the issue with objects defined in the `__main__` module.

        We need to handle the case where the object is defined in the `__main__` module.
        For some reason, this can lead to pickle issues.
        Based on some information I found, this should not happen, but it still does...
        To fix it, we detect, when an object is defined in `__main__` and temporarily add it to the "real" module
        representing the main function.
        Afterwards we do some cleanup.
        Not sure if really required, but it seems to work.
        Overall very hacky, but I don't see a better way to fix this.

        The implementation provided by the parent class (joblib.hashing.Hasher) should work, but it does not.
        I think the second call to `save_global` there is causing the issue.
        In our version we first fix the module before calling save_global again.

        Note that we are not using super() here, as we want to skip the faulty implementation in the parent class.

        We also need to modify the dispatch table to force our custom save_global method.
        See below this method for the implementation.
        """
        kwargs = {"name": name, "pack": pack}
        del kwargs["pack"]
        try:
            Pickler.save_global(self, obj, **kwargs)
        except pickle.PicklingError:
            modules_modified = []
            if getattr(obj, "__module__", None) == "__main__":
                try:
                    name = obj.__qualname__
                    to_add_obj = obj
                except AttributeError:
                    name = obj.__class__.__qualname__
                    to_add_obj = obj.__class__
                mod = sys.modules["__main__"]
                if not hasattr(mod, name):
                    modules_modified.append((mod, name))
                    setattr(mod, name, to_add_obj)
            try:
                Pickler.save_global(self, obj, **kwargs)
            finally:
                # Remove all new entries made to the main module.
                for mod, name in modules_modified:
                    delattr(mod, name)

    # We also need to rewrite the dispatch table to force our custom save_global method.
    dispatch = Hasher.dispatch.copy()
    # builtin
    dispatch[type(len)] = save_global
    # type
    dispatch[type(object)] = save_global
    # classobj
    dispatch[type(Hasher)] = save_global
    # function
    dispatch[type(pickle.dump)] = save_global


class NoMemoizeNumpyHasher(NoMemoizeHasher, NumpyHasher):
    """A joblib numpy hasher with all memoize features disabled."""


class NNHasher(NoMemoizeNumpyHasher):
    """A hasher that can handle torch models.

    Under the hood this uses the new implementation of `torch.save` to serialize the model.
    This produces consistent output.

    Note: I never did any performance checks with large models.
    """

    def __init__(self, hash_name="md5", coerce_mmap=False) -> None:
        super().__init__(hash_name, coerce_mmap)
        try:
            import torch  # pylint: disable=import-outside-toplevel

            self.torch = torch
        except ImportError:
            self.torch = None
        try:
            import tensorflow

            self.tensorflow = tensorflow
        except ImportError:
            self.tensorflow = None

    def save(self, obj):
        if self.torch and isinstance(obj, (self.torch.nn.Module, self.torch.Tensor)):
            b = b""
            buffer = io.BytesIO(b)
            self.torch.save(obj, buffer)
            self._hash.update(b)
            return

        if self.tensorflow and isinstance(obj, (self.tensorflow.keras.Model,)):
            # The normal tensorflow objects don't have a consistent hash.
            # Therefore, we need to serialize all relevant information.
            # I hope by using `serialize_keras_object` we get all important parts.
            # The `SharedObjectSavingScope` is required to make sure that shared objects are properly serialized.
            from tensorflow.python.keras.utils.generic_utils import SharedObjectSavingScope, serialize_keras_object

            with SharedObjectSavingScope():
                NumpyHasher.save(
                    self,
                    [obj.__class__.__name__, serialize_keras_object(obj), obj.get_weights()],
                )
            return
        NumpyHasher.save(self, obj)


# This function is modified based on
# https://github.com/joblib/joblib/blob/4dafaff788a3b5402acfed091558b4c511982959/joblib/hashing.py#L244
def custom_hash(obj, hash_name="md5", coerce_mmap=False):
    """Quick calculation of a hash to identify uniquely Python objects containing numpy arrays and torch models.

    This function is modified based on `joblib.hash` so that it can properly handle torch models.

    Parameters
    ----------
    obj
        The object to be hashed
    hash_name: 'md5' or 'sha1'
        Hashing algorithm used. sha1 is supposedly safer, but md5 is faster.
    coerce_mmap: boolean
        Make no difference between np.memmap and np.ndarray

    """
    valid_hash_names = ("md5", "sha1")
    if hash_name not in valid_hash_names:
        raise ValueError(f"Valid options for 'hash_name' are {valid_hash_names}. Got hash_name={hash_name!r} instead.")
    if "torch" in sys.modules or "tensorflow" in sys.modules:
        hasher = NNHasher(hash_name=hash_name, coerce_mmap=coerce_mmap)
    elif "numpy" in sys.modules:
        hasher = NoMemoizeNumpyHasher(hash_name=hash_name, coerce_mmap=coerce_mmap)
    else:
        hasher = NoMemoizeHasher(hash_name=hash_name)
    with Path(os.devnull).open("w") as devnull, contextlib.redirect_stdout(devnull):
        # Some object decide to print stuff to stdout when pickling.
        # As we potentially pickle a lot of objects, we don't want to see this.
        return hasher.hash(obj)



"""Base class for all datasets."""
import warnings
from collections.abc import Iterator, Sequence
from keyword import iskeyword
from typing import Optional, TypeVar, Union, cast, overload

import numpy as np
import pandas as pd
from typing_extensions import Self

from tpcp._base import BaseTpcpObject, _recursive_validate, get_param_names
from tpcp.exceptions import ValidationError

DatasetT = TypeVar("DatasetT", bound="_Dataset")


class _Dataset(BaseTpcpObject):
    groupby_cols: Optional[Union[list[str], str]]
    subset_index: Optional[pd.DataFrame]

    @property
    def index(self) -> pd.DataFrame:
        """Get index."""
        # Before we do anything, we call manually trigger validation.
        # This is usually done by the `get_params` call, but for dataset classes there is a high chance that people
        # use the class without ever (explicitly or implicitly) calling `get_params`.
        _recursive_validate(type(self))

        if self.subset_index is None:
            # Note, in the past we recreated the index when ever there was a call to `self.index` and their was no
            # subset index.
            # To avoid unnecessary overhead and potential edgecases where the index creation is not deterministic (
            # i.e. the data on disc changes between calls), we now only create the index once.
            # This still doesn't solve the issues of having a non-deterministic index creation (still bad!),
            # but the errors that you are running into are hopefully less obscure.
            self.subset_index = self._create_check_index()

        return self.subset_index

    @classmethod
    def __custom_tpcp_validation__(cls):
        """Run custom validation for dataset classes.

        We check that the subclass correctly implements `groupby_cols` and `subset_index`.
        """
        if cls is _Dataset or cls is Dataset:
            return
        super().__custom_tpcp_validation__()
        params = get_param_names(cls)
        if "groupby_cols" not in params or "subset_index" not in params:
            raise ValidationError(
                f"Dataset classes must implement `groupby_cols` and `subset_index` as parameters and forward them to "
                f"the `super().__init__` call.\n"
                f"{cls.__name__} does only implement the following parameters: {params}\n\n"
                "Add `groupby_cols: Optional[Union[List[str], str]] = None, subset_index: Optional[pd.DataFrame] = "
                "None` to the end of the input arguments in the `__init__` method and forward them to the "
                "`super().__init__` call."
            )

    def _create_check_index(self):
        """Check the index creation.

        We create the index twice to check if the index creation is deterministic.
        If not we raise an error.
        This is fundamentally important for datasets to be deterministic.
        While we can not catch all related issues (i.e. determinism across different machines), this should catch the
        most obvious ones.

        Furthermore, we check if all columns of the index are valid Python attribute names and throw a warning if not.

        In case, creating the index twice is too expensive, users can overwrite this method.
        But better to catch errors early.
        """
        index_1 = self.create_index()
        index_2 = self.create_index()
        if not index_1.equals(index_2):
            raise RuntimeError(
                "Index creation is not deterministic! "
                "This is a fundamental requirement for datasets to be deterministic otherwise you might run into all "
                "kinds of issues. "
                "Please check your implementation of the `create_index` method.\n\n"
                "Typically sources of non-determinism are:\n"
                " - Using `random` somewhere in the code\n"
                " - Storing (intermediate) data in non-sorted containers (e.g. `set`)\n"
                " - Relying on the ordering of files from the file system\n\n"
                "For the last to cases we recommend to sort the dataframe you return from `create_index` "
                "explicitly using `sort_values`."
            )

        invalid_elements = [s for s in index_1.columns if not s.isidentifier() or iskeyword(s)]
        if invalid_elements:
            warnings.warn(
                f"Some of your index columns are not valid Python attribute names: {invalid_elements}. "
                "This will cause issues when using further methods such as `get_subset`, `group_label`, "
                "`group_labels`, and `datapoint_label`.\n"
                "To be a valid Python attribute a string should only consist of alphanumeric letters and underscores. "
                "Strings starting with a number, containing spaces or special characters are not allowed."
                "Furthermore, they must not shadow a built-in Python keyword.",
                RuntimeWarning,
                stacklevel=1,
            )

        return index_1

    @property
    def group_labels(self) -> list[tuple[str, ...]]:
        """Get all group labels of the dataset based on the set groupby level.

        This will return a list of named tuples.
        The tuples will contain only one entry if there is only one groupby level or index column.

        The elements of the named tuples will have the same names as the groupby columns and will be in the same order.

        Note, that if one of the groupby levels/index columns is not a valid Python attribute name (e.g. in contains
        spaces or starts with a number), the named tuple will not contain the correct column name!
        For more information see the documentation of the `rename` parameter of :func:`collections.namedtuple`.

        For some examples and additional explanation see this :ref:`example <custom_dataset_basics>`.
        """
        return list(
            self._get_unique_groups().to_frame().itertuples(index=False, name=type(self).__name__ + "GroupLabel")
        )

    @property
    def groups(self) -> list[tuple[str, ...]]:
        """Get the current group labels. Deprecated, use `group_labels` instead."""
        warnings.warn(
            "The attribute `groups` is deprecated and will be removed in a future version. "
            "Use `group_labels` instead.",
            DeprecationWarning,
            stacklevel=1,
        )
        return self.group_labels

    @property
    def group_label(self) -> tuple[str, ...]:
        """Get the current group label.

        The group is defined by the current groupby settings.
        If the dataset is not grouped, this is equivalent to `datapoint_label`.

        Note, this attribute can only be used, if there is just a single group.
        This will return a named tuple. The tuple will contain only one entry if there is only a single groupby column
        or column in the index.
        The elements of the named tuple will have the same names as the groupby columns and will be in the same order.
        """
        self.assert_is_single_group("group")
        return self.group_labels[0]

    @property
    def group(self) -> tuple[str, ...]:
        """Get the current group label. Deprecated, use `group_label` instead."""
        warnings.warn(
            "The attribute `group` is deprecated and will be removed in a future version. "
            "Use `group_label` instead.",
            DeprecationWarning,
            stacklevel=1,
        )
        return self.group_label

    def index_as_tuples(self) -> list[tuple[str, ...]]:
        """Get all datapoint labels of the dataset (i.e. a list of the rows of the index as named tuples)."""
        return list(self.index.itertuples(index=False, name=type(self).__name__ + "DatapointLabel"))

    def __len__(self) -> int:
        """Get the length of the dataset.

        This is equal to the number of rows in the index, if `self.groupby_cols=None`.
        Otherwise, it is equal to the number of unique groups.
        """
        return len(self.group_labels)

    @property
    def shape(self) -> tuple[int]:
        """Get the shape of the dataset.

        This only reports a single dimension.
        This is equal to the number of rows in the index, if `self.groupby_cols=None`.
        Otherwise, it is equal to the number of unique groups.
        """
        return (len(self),)

    @property
    def grouped_index(self) -> pd.DataFrame:
        """Return the index with the `groupby` columns set as multiindex."""
        groupby_cols = self._get_groupby_columns()
        try:
            return self.index.set_index(groupby_cols, drop=False)
        except KeyError as e:
            raise KeyError(
                f"You can only groupby columns that are part of the index columns ({list(self.index.columns)}) and not"
                f" {self.groupby_cols}"
            ) from e

    def _get_groupby_columns(self) -> list[str]:
        """Get the groupby columns."""
        if self.groupby_cols is None:
            return self.index.columns.to_list()
        return _ensure_is_list(self.groupby_cols)

    def _get_unique_groups(self) -> Union[pd.MultiIndex, pd.Index]:
        return self.grouped_index.index.unique()

    def __getitem__(self, subscript: Union[int, Sequence[int], np.ndarray, slice]) -> Self:
        """Return a dataset object containing only the selected row indices of `self.group_labels`."""
        multi_index = self._get_unique_groups()[subscript]
        if not isinstance(multi_index, pd.Index):
            multi_index = [multi_index]

        return self.clone().set_params(subset_index=self.grouped_index.loc[multi_index].reset_index(drop=True))

    def groupby(self, groupby_cols: Optional[Union[list[str], str]]) -> Self:
        """Return a copy of the dataset grouped by the specified columns.

        This does not change the order of the rows of the dataset index.

        Each unique group represents a single data point in the resulting dataset.

        Parameters
        ----------
        groupby_cols
            None (no grouping) or a valid subset of the columns available in the dataset index.

        """
        grouped_ds = self.clone().set_params(groupby_cols=groupby_cols)
        # Get grouped index to raise an error here, in case the `groupby_cols` are invalid.
        _ = grouped_ds.grouped_index
        return grouped_ds

    def get_subset(
        self,
        *,
        group_labels: Optional[list[tuple[str, ...]]] = None,
        index: Optional[pd.DataFrame] = None,
        bool_map: Optional[Sequence[bool]] = None,
        **kwargs: Union[list[str], str],
    ) -> Self:
        """Get a subset of the dataset.

        .. note::
            All arguments are mutable exclusive!

        Parameters
        ----------
        group_labels
            A valid row locator or slice that can be passed to `self.grouped_index.loc[locator, :]`.
            This basically needs to be a subset of `self.group_labels`.
            Note that this is the only indexer that works on the grouped index.
            All other indexers work on the pure index.
        index
            `pd.DataFrame` that is a valid subset of the current dataset index.
        bool_map
            bool-map that is used to index the current index-dataframe.
            The list **must** be of same length as the number of rows in the index.
        **kwargs
            The key **must** be the name of an index column.
            The value is a list containing strings that correspond to the categories that should be kept.
            For examples see above.

        Returns
        -------
        subset
            New dataset object filtered by specified parameters.

        """
        if [x is None or (isinstance(x, dict) and len(x) == 0) for x in [group_labels, index, bool_map, kwargs]].count(
            False
        ) > 1:
            raise ValueError("Only one of `group_labels`, `selected_keys`, `index`, `bool_map` or kwarg can be set!")

        if group_labels is not None:
            return self.clone().set_params(subset_index=self.grouped_index.loc[group_labels, :].reset_index(drop=True))

        if index is not None:
            if len(index) == 0:
                raise ValueError("Provided index is not formatted correctly. Make sure it is not empty!")

            return self.clone().set_params(subset_index=index.reset_index(drop=True))

        if bool_map is not None:
            if len(bool_map) != self.index.shape[0]:
                raise ValueError(f"Parameter bool_map must have length {self.index.shape[0]} but has {len(bool_map)}!")

            return self.clone().set_params(subset_index=self.index[bool_map].reset_index(drop=True))

        if len(kwargs) > 0:
            cleaned_kwargs = cast(dict[str, list[str]], {k: _ensure_is_list(v) for k, v in kwargs.items()})

            # Check if all values are actually in their respective columns.
            # This is not strictly required, but avoids user error
            _assert_all_in_df(self.index, cleaned_kwargs)

            subset_index = self.index.loc[
                self.index[list(cleaned_kwargs.keys())].isin(cleaned_kwargs).all(axis=1)
            ].reset_index(drop=True)
            if len(subset_index) == 0:
                raise KeyError(f"No datapoint in the dataset matched the following filter: {cleaned_kwargs}")

            return self.clone().set_params(subset_index=subset_index)

        raise ValueError(
            "At least one of `group_labels`, `selected_keys`, `index`, `bool_map` or kwarg must not be None!"
        )

    def __repr__(self) -> str:
        """Return string representation of the dataset object."""
        repr_index = self.index if self.groupby_cols is None else self.grouped_index
        repr_index = str(repr_index).replace("\n", "\n   ")
        return f"{self.__class__.__name__} [{self.shape[0]} groups/rows]\n\n   {repr_index}\n\n   "[:-5]

    def _repr_html_(self) -> str:
        """Return html representation of the dataset object."""
        repr_index = self.index if self.groupby_cols is None else self.grouped_index

        representation = repr_index._repr_html_()

        # For type checking
        assert isinstance(representation, str)

        df_repr = (
            representation.replace("<div>", '<div style="margin-top: 0em">')
            .replace('<table border="1" class="dataframe"', '<table style="margin-left: 3em;"')
            .replace("<th>", '<th style="text-align: center;">')
            .replace("<td>", '<td style="text-align: center; padding-left: 2em; padding-right: 2em;">')
        )
        return (
            f'<h4 style="margin-bottom: 0.1em;">{self.__class__.__name__} [{self.shape[0]} groups/rows]</h3>\n'
            + df_repr
        )

    def __iter__(self) -> Iterator[Self]:
        """Return generator object containing a subset for every combination up to and including the selected level."""
        return (self.__getitem__(i) for i in range(self.shape[0]))

    def iter_level(self, level: str) -> Iterator[Self]:
        """Return generator object containing a subset for every category from the selected level.

        Parameters
        ----------
        level
            Optional `str` that sets the level which shall be used for iterating.
            This **must** be one of the columns names of the index.

        Returns
        -------
        subset
            New dataset object containing only one category in the specified `level`.

        """
        if level not in self.index.columns:
            raise ValueError(f"`level` must be one of {list(self.index.columns)}")

        return (self.get_subset(**{level: category}) for category in self.index[level].unique())

    def is_single(self, groupby_cols: Optional[Union[str, list[str]]]) -> bool:
        """Return True if index contains only one row/group with the given groupby settings.

        If `groupby_cols=None` this checks if there is only a single row left.
        If you want to check if there is only a single group within the current grouping, use `is_single_group` instead.

        Parameters
        ----------
        groupby_cols
            None (no grouping) or a valid subset of the columns available in the dataset index.

        """
        return len(self.groupby(groupby_cols)) == 1

    def is_single_group(self) -> bool:
        """Return True if index contains only one group."""
        return len(self) == 1

    def assert_is_single(self, groupby_cols: Optional[Union[str, list[str]]], property_name) -> None:
        """Raise error if index does contain more than one group/row with the given groupby settings.

        This should be used when implementing access to data values, which can only be accessed when only a single
        trail/participant/etc. exist in the dataset.

        Parameters
        ----------
        groupby_cols
            None (no grouping) or a valid subset of the columns available in the dataset index.
        property_name
            Name of the property this check is used in.
            Used to format the error message.

        """
        if not self.is_single(groupby_cols):
            if groupby_cols is None:
                groupby_cols = self.index.columns.to_list()
            raise ValueError(
                f"The attribute `{property_name}` of dataset {self.__class__.__name__} can only be accessed if there "
                f"is only a single combination of the columns {groupby_cols} left in a data subset,"
            )

    def assert_is_single_group(self, property_name) -> None:
        """Raise error if index does contain more than one group/row.

        Note that this is different from `assert_is_single` as it is aware of the current grouping.
        Instead of checking that a certain combination of columns is left in the dataset, it checks that only a
        single group exists with the already selected grouping as defined by `self.groupby_cols`.

        Parameters
        ----------
        property_name
            Name of the property this check is used in.
            Used to format the error message.

        """
        if not self.is_single_group():
            if self.groupby_cols is None:
                group_error_str = (
                    "Currently the dataset is not grouped. "
                    "This means a single group is identical to having only a single row in the "
                    "dataset index."
                )
            else:
                group_error_str = f"Currently the dataset is grouped by {self.groupby_cols}."

            raise ValueError(
                f"The attribute `{property_name}` of dataset {self.__class__.__name__} can only be accessed if there is"
                f" only a single group left in a data subset. " + group_error_str
            )

    def create_group_labels(self, label_cols: Union[str, list[str]]) -> list[str]:
        warnings.warn(
            "The method `create_string_group_labels` is deprecated and will be removed in a future version. "
            "Use `create_string_group_labels` instead.",
            DeprecationWarning,
            stacklevel=1,
        )
        return self.create_string_group_labels(label_cols)

    def create_string_group_labels(self, label_cols: Union[str, list[str]]) -> list[str]:
        """Generate a list of string labels for each group/row in the dataset.

        .. note::
            This has a different use case than the dataset-wide groupby.
            Using `groupby` reduces the effective size of the dataset to the number of groups.
            This method produces a group label for each group/row that is already in the dataset, without changing the
            dataset.

        The output of this method can be used in combination with :class:`~sklearn.model_selection.GroupKFold` as
        the group label.

        Parameters
        ----------
        label_cols
            The columns that should be included in the label.
            If the dataset is already grouped, this must be a subset of  `self.groupby_cols`.

        """
        unique_index = self._get_unique_groups().to_frame()
        try:
            return [str(g) for g in unique_index.set_index(label_cols).index.to_list()]
        except KeyError as e:
            if self.groupby_cols is not None:
                raise KeyError(
                    "When using `create_string_group_labels` with a grouped dataset, the selected columns must "
                    f"be a subset of `self.groupby_cols` ({self.groupby_cols}) and not ({label_cols})"
                ) from e
            raise KeyError(
                f"The selected label columns ({label_cols}) are not in the index of the dataset "
                f"({list(self.index.columns)})."
            ) from e

    def create_index(self) -> pd.DataFrame:
        """Create the full index for the dataset.

        This needs to be implemented by the subclass.

        .. warning:: Make absolutely sure that the dataframe you return is deterministic and does not change between
                     runs!
                     This can lead to some nasty bugs!
                     We try to catch them internally, but it is not always possible.
                     As tips, avoid reliance on random numbers and make sure that the order is not depend on things
                     like file system order, when creating an index by scanning a directory.
                     Particularly nasty are cases when using non-sorted container like `set`, that sometimes maintain
                     their order, but sometimes don't.
                     At the very least, we recommend to sort the final dataframe you return in `create_index`.

        """
        raise NotImplementedError()


class Dataset(_Dataset):
    """Baseclass for tpcp Dataset objects.

    This class provides fundamental functionality like iteration, getting subsets, and compatibility with `sklearn`'s
    cross validation helpers.

    For more information check out the examples and user guides on datasets.

    Parameters
    ----------
    groupby_cols
        A column name or a list of column names that should be used to group the index before iterating over it.
        For examples see below.
    subset_index
        For all classes that inherit from this class, subset_index **must** be None by default.
        But the subclasses require a `create_index` method that returns a DataFrame representing the index.

    Examples
    --------
    This class is usually not meant to be used directly, but the following code snippets show some common operations
    that can be expected to work for all dataset subclasses.

    >>> import pandas as pd
    >>> from itertools import product
    >>>
    >>> from tpcp import Dataset
    >>>
    >>> test_index = pd.DataFrame(
    ...     list(product(("patient_1", "patient_2", "patient_3"), ("test_1", "test_2"), ("1", "2"))),
    ...     columns=["patient", "test", "extra"],
    ... )
    >>> # We create a little dummy dataset by passing an index directly to `test_index`
    >>> # Usually we would create a subclass with a `create_index` method that returns a DataFrame representing the
    >>> # index.
    >>> dataset = Dataset(subset_index=test_index)
    >>> dataset
    Dataset [12 groups/rows]
    <BLANKLINE>
             patient    test extra
       0   patient_1  test_1     1
       1   patient_1  test_1     2
       2   patient_1  test_2     1
       3   patient_1  test_2     2
       4   patient_2  test_1     1
       5   patient_2  test_1     2
       6   patient_2  test_2     1
       7   patient_2  test_2     2
       8   patient_3  test_1     1
       9   patient_3  test_1     2
       10  patient_3  test_2     1
       11  patient_3  test_2     2

    We can loop over the dataset.
    By default, we will loop over each row.

    >>> for r in dataset[:2]:
    ...     print(r)
    Dataset [1 groups/rows]
    <BLANKLINE>
            patient    test extra
       0  patient_1  test_1     1
    Dataset [1 groups/rows]
    <BLANKLINE>
            patient    test extra
       0  patient_1  test_1     2

    We can also change `groupby` (either in the init or afterwards), to loop over other combinations.
    If we select the level `test`, we will loop over all `patient`-`test` combinations.

    >>> grouped_dataset = dataset.groupby(["patient", "test"])
    >>> grouped_dataset  # doctest: +NORMALIZE_WHITESPACE
    Dataset [6 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_1 test_1  patient_1  test_1     1
                 test_1  patient_1  test_1     2
                 test_2  patient_1  test_2     1
                 test_2  patient_1  test_2     2
       patient_2 test_1  patient_2  test_1     1
                 test_1  patient_2  test_1     2
                 test_2  patient_2  test_2     1
                 test_2  patient_2  test_2     2
       patient_3 test_1  patient_3  test_1     1
                 test_1  patient_3  test_1     2
                 test_2  patient_3  test_2     1
                 test_2  patient_3  test_2     2

    >>> for r in grouped_dataset[:2]:
    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
    Dataset [1 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_1 test_1  patient_1  test_1     1
                 test_1  patient_1  test_1     2
    Dataset [1 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_1 test_2  patient_1  test_2     1
                 test_2  patient_1  test_2     2

    To iterate over the unique values of a specific level use the "iter_level" function:

    >>> for r in list(grouped_dataset.iter_level("patient"))[:2]:
    ...     print(r)  # doctest: +NORMALIZE_WHITESPACE
    Dataset [2 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_1 test_1  patient_1  test_1     1
                 test_1  patient_1  test_1     2
                 test_2  patient_1  test_2     1
                 test_2  patient_1  test_2     2
    Dataset [2 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_2 test_1  patient_2  test_1     1
                 test_1  patient_2  test_1     2
                 test_2  patient_2  test_2     1
                 test_2  patient_2  test_2     2

    We can also get arbitary subsets from the dataset:

    >>> subset = grouped_dataset.get_subset(patient=["patient_1", "patient_2"], extra="2")
    >>> subset  # doctest: +NORMALIZE_WHITESPACE
    Dataset [4 groups/rows]
    <BLANKLINE>
                           patient    test extra
       patient   test
       patient_1 test_1  patient_1  test_1     2
                 test_2  patient_1  test_2     2
       patient_2 test_1  patient_2  test_1     2
                 test_2  patient_2  test_2     2

    If we want to use datasets in combination with :class:`~sklearn.model_selection.GroupKFold`, we can generate
    valid group labels as follows.
    These grouplabels are strings representing the unique value of the index at the specified levels.

    .. note::
        You usually don't want to use that in combination with `self.groupby`.

    >>> # We are using the ungrouped dataset again!
    >>> group_labels = dataset.create_string_group_labels(["patient", "test"])
    >>> pd.concat([dataset.index, pd.Series(group_labels, name="group_labels")], axis=1)
          patient    test extra             group_labels
    0   patient_1  test_1     1  ('patient_1', 'test_1')
    1   patient_1  test_1     2  ('patient_1', 'test_1')
    2   patient_1  test_2     1  ('patient_1', 'test_2')
    3   patient_1  test_2     2  ('patient_1', 'test_2')
    4   patient_2  test_1     1  ('patient_2', 'test_1')
    5   patient_2  test_1     2  ('patient_2', 'test_1')
    6   patient_2  test_2     1  ('patient_2', 'test_2')
    7   patient_2  test_2     2  ('patient_2', 'test_2')
    8   patient_3  test_1     1  ('patient_3', 'test_1')
    9   patient_3  test_1     2  ('patient_3', 'test_1')
    10  patient_3  test_2     1  ('patient_3', 'test_2')
    11  patient_3  test_2     2  ('patient_3', 'test_2')

    """

    def __init__(
        self,
        *,
        groupby_cols: Optional[Union[list[str], str]] = None,
        subset_index: Optional[pd.DataFrame] = None,
    ) -> None:
        self.groupby_cols = groupby_cols
        self.subset_index = subset_index

    @classmethod
    def as_dataclass(cls):
        """Return a version of the Dataset class that can be subclassed using dataclasses."""
        # You are only allowed to call this method on the base class not the subclass!
        assert cls is Dataset, "You can only call `as_dataclass` on the base class `Dataset`!"

        import dataclasses  # pylint: disable=import-outside-toplevel

        @dataclasses.dataclass(eq=False, repr=False, order=False)
        class DatasetDc(_Dataset):
            """Dataclass version of Dataset."""

            groupby_cols: Optional[Union[list[str], str]] = None
            subset_index: Optional[pd.DataFrame] = None

        return DatasetDc

    @classmethod
    def as_attrs(cls):
        """Return a version of the Dataset class that can be subclassed using `attrs` defined classes.

        Note, this requires `attrs` to be installed!
        """
        assert cls is Dataset, "You can only call `as_attrs` on the base class `Dataset`!"

        from attrs import define  # pylint: disable=import-outside-toplevel

        @define(eq=False, repr=False, order=False, kw_only=True, slots=False)
        class DatasetAt(_Dataset):
            """Attrs version of Dataset."""

            groupby_cols: Optional[Union[list[str], str]] = None
            subset_index: Optional[pd.DataFrame] = None

        return DatasetAt


T = TypeVar("T")


@overload
def _ensure_is_list(x: list[T]) -> list[T]:
    ...


@overload
def _ensure_is_list(x: T) -> list[T]:
    ...


def _ensure_is_list(x):
    return x if isinstance(x, list) else [x]


def _assert_all_in_df(df, dic):
    """Check that all values of the dictionary are in the column 'key' of the pandas dataframe."""
    for key, value in dic.items():
        try:
            index_level = df[key]
        except KeyError as e:
            raise KeyError(f"Can not filter by key `{key}`! Key must be one of {list(df.columns)}!") from e
        if not set(value).issubset(index_level):
            raise KeyError(f"At least one of {value} is not in level {key}")



"""tpcp - Tiny Pipelines for Complicated Problems."""
from tpcp._algorithm import Algorithm
from tpcp._algorithm_utils import (
    get_action_method,
    get_action_methods_names,
    get_action_params,
    get_results,
    is_action_applied,
    make_action_safe,
    make_optimize_safe,
)
from tpcp._base import NOTHING, BaseFactory, BaseTpcpObject, CloneFactory, cf, clone, get_param_names
from tpcp._dataset import Dataset
from tpcp._parameters import (
    HyperPara,
    HyperParameter,
    OptimizableParameter,
    OptiPara,
    Para,
    Parameter,
    PurePara,
    PureParameter,
)
from tpcp._pipeline import OptimizablePipeline, Pipeline

__version__ = "0.29.0"


__all__ = [
    "make_action_safe",
    "make_optimize_safe",
    "clone",
    "cf",
    "CloneFactory",
    "BaseFactory",
    "HyperParameter",
    "HyperPara",
    "PureParameter",
    "PurePara",
    "OptimizableParameter",
    "OptiPara",
    "Parameter",
    "Para",
    "Algorithm",
    "BaseTpcpObject",
    "Dataset",
    "Pipeline",
    "OptimizablePipeline",
    "get_param_names",
    "get_action_params",
    "get_results",
    "get_action_method",
    "get_action_methods_names",
    "is_action_applied",
    "NOTHING",
]



import warnings
from collections.abc import Iterable, Iterator, Sequence
from dataclasses import fields, is_dataclass
from typing import Any, Callable, Generic, TypeVar

from tpcp import Algorithm, cf

DataclassT = TypeVar("DataclassT")
T = TypeVar("T")


class _NotSet:
    def __repr__(self):
        return "_NOT_SET"


class BaseTypedIterator(Algorithm, Generic[DataclassT]):
    """A Base class to implement custom typed iterators.

    This class is missing the ``iterate`` method, which needs to be implemented by the child class.
    It has a `_iterate` method though that does most of the heavy lifting.
    The actual iterate method should handle turning the inputs into an iterable and then call `_iterate`.

    :class:`TypedIterator` provides a "dummy" implementation that expects any type of iterator for the iterate method.
    Custom base classes could provide more elaborate preprocessing of the inputs before iteration.
    For example, cutting sections out of a dataframe based on a list of start and end indices, and then iterating over
    the cut sections.

    Parameters
    ----------
    data_type
        A dataclass that defines the result type you expect from each iteration.
    aggregations
        An optional list of aggregations to apply to the results.
        This has the form ``[(result_name, aggregation_function), ...]``.
        If a result-name is in the list, the aggregation will be applied to it, when accessing the respective result
        attribute (i.e. ``{result_name}_``).
        If no aggregation is defined for a result, a simple list of all results will be returned.
    NULL_VALUE
        (Class attribute) The value that is used to initialize the result dataclass and will remain in the results, if
        no result was for a specific attribute in one or more iterations.

    Attributes
    ----------
    inputs_
        List of all input elements that were iterated over.
    raw_results_
        List of all results as dataclass instances.
        The attribute of the dataclass instance will have the a value of ``_NOT_SET`` if no result was set.
        To check for this, you can use ``isinstance(val, TypedIterator.NULL_VALUE)``.
    {result_name}_
        The aggregated results for the respective result name.
    done_
        True, if the iterator is done.
        If the iterator is not done, but you try to access the results, a warning will be raised.

    """

    data_type: type[DataclassT]
    aggregations: Sequence[tuple[str, Callable[[list, list], Any]]]

    _raw_results: list[DataclassT]
    done_: bool
    inputs_: list

    NULL_VALUE = _NotSet()

    def __init__(
        self, data_type: type[DataclassT], aggregations: Sequence[tuple[str, Callable[[list, list], Any]]] = cf([])
    ):
        self.data_type = data_type
        self.aggregations = aggregations

    def _iterate(self, iterable: Iterable[T]) -> Iterator[tuple[T, DataclassT]]:
        """Iterate over the given iterable and yield the input and a new empty result object for each iteration.

        Parameters
        ----------
        iterable
            The iterable to iterate over.

        Yields
        ------
        input, result_object
            The input and a new empty result object.
            The result object is a dataclass instance of the type defined in ``self.data_type``.
            All values of the result object are set to ``TypedIterator.NULL_VALUE`` by default.

        """
        if not is_dataclass(self.data_type):
            raise TypeError(f"Expected a dataclass as data_type, got {self.data_type}")

        self._raw_results = []
        self.inputs_ = []
        self.done_ = False
        for d in iterable:
            result_object = self._get_new_empty_object()
            self._raw_results.append(result_object)
            self.inputs_.append(d)
            yield d, result_object
        self.done_ = True

    def _get_new_empty_object(self) -> DataclassT:
        init_dict = {k.name: self.NULL_VALUE for k in fields(self.data_type)}
        return self.data_type(**init_dict)

    @property
    def raw_results_(self) -> list[DataclassT]:
        if not self.done_:
            warnings.warn("The iterator is not done yet. The results might not be complete.", stacklevel=1)

        return self._raw_results

    def __getattr__(self, item):
        # We assume a correct result name ends with an underscore
        actual_item = item[:-1]

        if actual_item in self._raw_results[0].__dict__:
            values = [getattr(r, actual_item) for r in self.raw_results_]
            # if an aggregator is defined for the specific item, we apply it
            for name, aggregator in self.aggregations:
                if name == actual_item:
                    return aggregator(self.inputs_, values)
            return values

        valid_result_fields = [k.name + "_" for k in fields(self.data_type)]

        raise AttributeError(
            f"Attribute {item} is not a valid attribute for {self.__class__.__name__} nor a dynamically generated "
            "result attribute of the result dataclass. "
            f"Valid result attributes are: {valid_result_fields}. "
            "Note the trailing underscore!"
        )


class TypedIterator(BaseTypedIterator, Generic[DataclassT]):
    """Helper to iterate over data and collect results.

    Parameters
    ----------
    data_type
        A dataclass that defines the result type you expect from each iteration.
    aggregations
        An optional list of aggregations to apply to the results.
        This has the form ``[(result_name, aggregation_function), ...]``.
        If a result-name is in the list, the aggregation will be applied to it, when accessing the respective result
        attribute (i.e. ``{result_name}_``).
        If no aggregation is defined for a result, a simple list of all results will be returned.
    NULL_VALUE
        (Class attribute) The value that is used to initialize the result dataclass and will remain in the results, if
        no result was for a specific attribute in one or more iterations.

    Attributes
    ----------
    inputs_
        List of all input elements that were iterated over.
    raw_results_
        List of all results as dataclass instances.
        The attribute of the dataclass instance will have the value of ``_NOT_SET`` if no result was set.
        To check for this, you can use ``isinstance(val, TypedIterator.NULL_VALUE)``.
    {result_name}_
        The aggregated results for the respective result name.
    done_
        True, if the iterator is done.
        If the iterator is not done, but you try to access the results, a warning will be raised.

    """

    data_type: type[DataclassT]
    aggregations: Sequence[tuple[str, Callable[[list, list], Any]]]

    _raw_results: list[DataclassT]
    done_: bool
    inputs_: list

    NULL_VALUE = _NotSet()

    def iterate(self, iterable: Iterable[T]) -> Iterator[tuple[T, DataclassT]]:
        """Iterate over the given iterable and yield the input and a new empty result object for each iteration.

        Parameters
        ----------
        iterable
            The iterable to iterate over.

        Yields
        ------
        input, result_object
            The input and a new empty result object.
            The result object is a dataclass instance of the type defined in ``self.data_type``.
            All values of the result object are set to ``TypedIterator.NULL_VALUE`` by default.

        """
        yield from self._iterate(iterable)



"""A set of helper functions to correctly handle global variables in parallel processes.

Both functions are required as long as https://github.com/joblib/joblib/issues/1071 is not resolved.

The provided workarounds are similar to the ones done in scikit-learn
(https://github.com/scikit-learn/scikit-learn/pull/25363).

Note, that these fixes are not necessarily compatible with each other.
This means you can not forward custom callbacks through scikit-learn Parallel calls.
The same way, by default tpcp will not forward sklearn global configs through tpcp Parallel calls.
However, you can likely configure the callbacks in tpcp to make that work.
"""
import functools
from typing import Callable, TypeVar

import joblib

T = TypeVar("T")
CalbackReturnType = tuple[T, Callable[[T], None]]
_PARALLEL_CONTEXT_CALLBACKS: list[Callable[[], CalbackReturnType]] = []


def delayed(func):
    """Wrap a function to be used in a parallel context.

    This is a modified version of joblib.delayed that can run arbitrary callbacks when `delayed` is called in the
    main process and when the function is called in the parallel process.

    This is useful for example to restore the global config in the parallel process.
    For this to work, callbacks must be registered using :func:`~tpcp.parallel.register_global_parallel_callback` first.

    All uses of `delayed` in tpcp are using this implementation.
    This means you can configure your custom callbacks and expect them to work in all tpcp functions that use
    multiprocessing.

    If you need to write your own multiprocessing method using joblib, refer to the example below.

    Examples
    --------
    This example shows how to use this to make sure the global scikit-learn config is restored in the parallel process
    used in tpcp.
    Note, sklearn has a custom workaround for this, which is not compatible with tpcp.

    >>> from tpcp.parallel import delayed, register_global_parallel_callback
    >>> from joblib import Parallel
    >>> from sklearn import get_config, set_config
    >>>
    >>> set_config(assume_finite=True)
    >>> def callback():
    ...     def setter(config):
    ...         set_config(**config)
    ...
    ...     return get_config(), setter
    >>>
    >>> def worker_func():
    ...     # This is what would be called in the parallel process
    ...     # We just return the config here for demonstration purposes
    ...     config = get_config()
    ...     return config["assume_finite"]
    >>>
    >>> # register the callback
    >>> register_global_parallel_callback(callback)
    >>> # call the worker function in parallel
    >>> Parallel(n_jobs=2)(delayed(worker_func)() for _ in range(2))
    [True, True]

    Notes
    -----
    The getters are called as soon as the delayed function is called in the main process.
    This means, if you are calling `delayed` long before the actual parallel execution, the getters might not capture
    the correct state of the global variables.

    """
    _parallel_setter = []
    for g in _PARALLEL_CONTEXT_CALLBACKS:
        _parallel_setter.append(g())

    @functools.wraps(func)
    def inner(*args, **kwargs):
        for value, setter in _parallel_setter:
            setter(value)
        return func(*args, **kwargs)

    return joblib.delayed(inner)


def register_global_parallel_callback(callback: Callable[[], tuple[T, Callable[[T], None]]]):
    """Register a callback to transfer information to parallel workers.

    This callback should be used together with :func:`~tpcp.parallel.delayed` in a joblib parallel context.
    All calbacks that are registered here will be called when :func:`~tpcp.parallel.delayed` is called in the main
    process.
    We expect that the callback returns a value and a setter function.
    In the worker process, the setter function will be called with the value that was returned by the callback.

    As all tpcp functions that use multiprocessing are using :func:`~tpcp.parallel.delayed`, registered callbacks will
    be called correctly in all tpcp functions.

    For details see the docs of :func:`~tpcp.parallel.delayed`.

    Parameters
    ----------
    callback : Callable[[], Tuple[T, Callable[[T], None]]]
        The callback function that will be called when :func:`~tpcp.parallel.delayed` is called in the main process.
        The callback should return a value and a setter function.

    """
    _PARALLEL_CONTEXT_CALLBACKS.append(callback)


__all__ = ["delayed", "register_global_parallel_callback"]



"""Helper for testing of algorithms and pipelines implemented in tpcp."""
import pytest

pytest.register_assert_rewrite("tpcp.testing._algorithm_test_mixin")
pytest.register_assert_rewrite("tpcp.testing._regression_utils")

from tpcp.testing._algorithm_test_mixin import TestAlgorithmMixin  # noqa: E402
from tpcp.testing._regression_utils import PyTestSnapshotTest  # noqa: E402

__all__ = ["TestAlgorithmMixin", "PyTestSnapshotTest"]



"""Helper to score pipelines."""
from __future__ import annotations

import traceback
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
)

import numpy as np
from joblib import Parallel
from tqdm import tqdm
from typing_extensions import Protocol

from tpcp import NOTHING
from tpcp._base import BaseTpcpObject, _Nothing
from tpcp._dataset import Dataset, DatasetT
from tpcp._pipeline import Pipeline, PipelineT
from tpcp._utils._general import _passthrough
from tpcp.exceptions import ScorerFailedError, ValidationError
from tpcp.parallel import delayed

if TYPE_CHECKING:
    from collections.abc import Sequence

T = TypeVar("T")
AggReturnType = Union[float, dict[str, float], _Nothing]

SingleScoreTypeT = Union[T, "Aggregator[Any]"]
MultiScoreTypeT = dict[str, SingleScoreTypeT[T]]
ScoreTypeT = Union[SingleScoreTypeT[T], MultiScoreTypeT[T]]

ScoreFuncSingle = Callable[[PipelineT, DatasetT], SingleScoreTypeT[T]]
ScoreFuncMultiple = Callable[[PipelineT, DatasetT], MultiScoreTypeT[T]]
ScoreFunc = Callable[[PipelineT, DatasetT], ScoreTypeT[T]]


class ScoreCallback(Protocol[PipelineT, DatasetT, T]):
    """Callback signature for scorer callbacks."""

    def __call__(
        self,
        *,
        step: int,
        scores: tuple[ScoreTypeT[T], ...],
        scorer: Scorer[PipelineT, DatasetT, T],
        pipeline: PipelineT,
        dataset: DatasetT,
    ) -> None:
        ...


class Aggregator(Generic[T]):
    """Base class for aggregators.

    You can subclass this class to create your own aggregators.
    The only thing you should change, is to overwrite the `aggregate` method.
    Everything else should not be modified.

    Custom aggregators can then be used to wrap return values of score functions or they can be passed as
    `default_aggregator` to the :class:`~tpcp.validate.Scorer` class.
    """

    _value: T

    RETURN_RAW_SCORES: ClassVar[bool] = True

    def __init__(self, _value: T) -> None:
        self._value = _value

    def __repr__(self) -> str:
        """Show the representation of the object."""
        return f"{self.__class__.__name__}({self._value!r})"

    def get_value(self) -> T:
        """Return the value wrapped by aggregator."""
        return self._value

    @classmethod
    def aggregate(cls, /, values: Sequence[T], datapoints: Sequence[Dataset]) -> AggReturnType:
        """Aggregate the values."""
        raise NotImplementedError()


class MeanAggregator(Aggregator[float]):
    """Aggregator that calculates the mean of the values."""

    @classmethod
    def aggregate(cls, /, values: Sequence[float], datapoints: Sequence[Dataset]) -> float:  # noqa: ARG003
        """Aggregate a sequence of floats by taking the mean."""
        try:
            return float(np.mean(values))
        except TypeError as e:
            raise ValidationError(
                f"MeanAggregator can only be used with float values. Got the following values instead:\n\n{values}"
            ) from e


class NoAgg(Aggregator[Any]):
    """Wrapper to wrap one or multiple output values of a scorer to prevent aggregation of these values.

    If one of the values in the return dictionary of a multi-value score function is wrapped with this class,
    the scorer will not aggregate the value.
    This allows to pass arbitary data from the score function through the scorer.
    As example, this could be some general metadata, some non-numeric scores, or an array of values (e.g. when the
    actual score is the mean of such values).

    Examples
    --------
    >>> def score_func(pipe, dataset):
    ...     ...
    ...     return {"score_val_1": score, "some_metadata": NoAgg(metadata)}
    >>> my_scorer = Scorer(score_func)

    """

    @classmethod
    def aggregate(cls, /, values: Sequence[Any], datapoints: Sequence[Dataset]) -> _Nothing:  # noqa: ARG003
        """Return nothing, indicating no aggregation."""
        return NOTHING


class Scorer(Generic[PipelineT, DatasetT, T], BaseTpcpObject):
    """A scorer to score multiple data points of a dataset and average the results.

    Parameters
    ----------
    score_func
        The callable that is used to score each data point
    single_score_callback
        Callback function that is called after each datapoint that is scored.
        It should have the following call signature:

        >>> def callback(
        ...     *, step: int, scores: Tuple[_SCORE_TYPE, ...], scorer: "Scorer", pipeline: Pipeline, dataset: Dataset, **_
        ... ) -> None:
        ...     ...

        All parameters will be passed as keyword arguments.
        This means, if your callback only needs a subset of the defined parameters, you can ignore them by using
        unused kwargs:

        >>> def callback(*, step: int, pipeline: Pipeline, **_):
        ...     ...

    n_jobs
        The number of parallel jobs to run.
        Each job will run on a single data point.
        Note, that the single_score_callback will still be called in the main thread, after a job is finished.
        However, it could be that multiple jobs are finished before the callback is called.
        The callback is still gurateed to be called in the order of the data points.
        If None, no parallelization is used.
    verbose
        Controls the verbosity of the parallelization.
        See :class:`joblib.Parallel` for more details.
    pre_dispatch
        Controls the number of jobs that get dispatched during parallelization.
        See :class:`joblib.Parallel` for more details.
    progress_bar
        True/False to enable/disable a `tqdm` progress bar.

    """

    score_func: ScoreFunc[PipelineT, DatasetT, T]
    default_aggregator: type[Aggregator[T]]
    single_score_callback: Optional[ScoreCallback[PipelineT, DatasetT, T]]
    n_jobs: Optional[int]
    verbose: int
    pre_dispatch: Union[str, int]
    progress_bar: bool

    def __init__(
        self,
        score_func: ScoreFunc[PipelineT, DatasetT, ScoreTypeT[T]],
        *,
        default_aggregator: type[Aggregator[T]] = MeanAggregator,
        single_score_callback: Optional[ScoreCallback[PipelineT, DatasetT, T]] = None,
        # Multiprocess_kwargs
        n_jobs: Optional[int] = None,
        verbose: int = 0,
        pre_dispatch: Union[str, int] = "2*n_jobs",
        progress_bar: bool = True,
    ) -> None:
        self.score_func = score_func
        self.default_aggregator = default_aggregator
        self.single_score_callback = single_score_callback
        self.n_jobs = n_jobs
        self.verbose = verbose
        self.pre_dispatch = pre_dispatch
        self.progress_bar = progress_bar

    # The typing for IndividualScoreType here is not perfect, but not sure how to fix.
    # For the aggregated scores, we can easily parameterize the value based on the generic, but not for the single
    # scores
    def __call__(
        self, pipeline: PipelineT, dataset: DatasetT
    ) -> tuple[Union[float, dict[str, float]], Union[Optional[list[T]], dict[str, list[T]]]]:
        """Score the pipeline with the provided data.

        Returns
        -------
        agg_scores
            The average scores over all data-points
        single_scores
            The scores for each individual data-point

        """
        return self._score(pipeline=pipeline, dataset=dataset)

    def _aggregate(  # noqa: C901, PLR0912
        self,
        scores: Union[tuple[type[Aggregator[T]], list[T]], dict[str, tuple[type[Aggregator[T]], list[T]]]],
        datapoints: list[DatasetT],
    ) -> tuple[Union[float, dict[str, float]], Union[Optional[list[T]], dict[str, list[T]]]]:
        if not isinstance(scores, dict):
            aggregator_single, raw_scores_single = scores
            if aggregator_single is NoAgg:
                raise ValidationError(
                    "Scorer returned a NoAgg object. "
                    "This is not allowed when returning only a single score value. "
                    "If you want to use a NoAgg scorer, return a dictionary of values, where one or "
                    "multiple values are wrapped with NoAgg."
                )
            # We create an instance of the aggregator here, even though we only need to call the class method.
            # This way, `aggregate` will work, even if people forgot to implement the aggregate method as class
            # method on their custom aggregator.
            agg_val = aggregator_single(None).aggregate(values=raw_scores_single, datapoints=datapoints)
            if isinstance(agg_val, dict):
                if not all(isinstance(score, float) for score in agg_val.values()):
                    raise ValidationError(
                        "Final aggregated scores are not all numbers. "
                        "Double-check your (custom) aggregators."
                        f"The current values are: \n{agg_val}"
                    )
            elif not isinstance(agg_val, (int, float)):
                raise ValidationError(f"The final aggregated score must be a numbers. Instead it is:\n{agg_val}")
            if aggregator_single.RETURN_RAW_SCORES is False:
                return agg_val, NOTHING
            return agg_val, list(raw_scores_single)

        raw_scores: dict[str, list[T]] = {}
        agg_scores: dict[str, float] = {}
        for name, (aggregator, raw_score) in scores.items():
            if aggregator.RETURN_RAW_SCORES is True:
                raw_scores[name] = list(raw_score)
            agg_score = aggregator(None).aggregate(values=raw_score, datapoints=datapoints)
            if agg_score is NOTHING or isinstance(agg_score, _Nothing):
                # This is the case with the NoAgg Scorer
                continue
            if isinstance(agg_score, dict):
                # If the aggregator returned multiple values, we merge them prefixing the original name
                for key, value in agg_score.items():
                    agg_scores[f"{name}__{key}"] = value
            else:
                agg_scores[name] = agg_score
        # Finally we check that all aggregates values are floats
        if not all(isinstance(score, (int, float)) for score in agg_scores.values()):
            raise ValidationError(
                "Final aggregated scores are not all numbers. "
                "Double-check your (custom) aggregators."
                f"The current values are:\n{agg_scores}"
            )
        return agg_scores, raw_scores

    def _score(self, pipeline: PipelineT, dataset: DatasetT):
        def per_datapoint(i, d):
            try:
                # We need to clone here again, to make sure that the run for each data point is truly independent.
                score = self.score_func(pipeline.clone(), d)
            except Exception as e:  # noqa: BLE001
                raise ScorerFailedError(
                    f"Scorer raised an exception while scoring data point {i} ({d.group_label}). "
                    "Tpcp does not support that (compared to sklearn) and you need to handle error cases yourself "
                    "within the scoring function."
                    "\n\n"
                    "The original exception was:\n\n"
                    f"{traceback.format_exc()}"
                ) from e
            return i, score

        scores: list[ScoreTypeT[T]] = []

        pbar = partial(tqdm, total=len(dataset), desc="Datapoints") if self.progress_bar else _passthrough
        parallel = Parallel(
            n_jobs=self.n_jobs, verbose=self.verbose, pre_dispatch=self.pre_dispatch, return_as="generator"
        )
        with parallel:
            for i, r in pbar(parallel(delayed(per_datapoint)(i, d) for i, d in enumerate(dataset))):
                scores.append(r)
                if self.single_score_callback:
                    self.single_score_callback(
                        step=i,
                        scores=tuple(scores),
                        scorer=self,
                        pipeline=pipeline,
                        dataset=dataset,
                    )

        return self._aggregate(_check_and_invert_score_dict(scores, self.default_aggregator), list(dataset))


ScorerTypes = Union[ScoreFunc[PipelineT, DatasetT, ScoreTypeT[T]], Scorer[PipelineT, DatasetT, ScoreTypeT[T]], None]


def _passthrough_scoring(pipeline: Pipeline[DatasetT], datapoint: DatasetT):
    """Call the score method of the pipeline to score the input."""
    return pipeline.score(datapoint)


def _validate_scorer(
    scoring: ScorerTypes[PipelineT, DatasetT, Any],
    pipeline: PipelineT,
    base_class: type[Scorer[Any, Any, Any]] = Scorer,
) -> Scorer[PipelineT, DatasetT, Any]:
    """Convert the provided scoring method into a valid scorer object."""
    if scoring is None:
        # If scoring is None, we will try to use the score method of the pipeline
        # However, we run score once with an empty dataset and check if it is actually implemented:
        try:
            pipeline.score(Dataset())  # type: ignore  # noqa: PGH003
        except NotImplementedError as e:
            raise e  # noqa: TRY201
        except Exception:  # noqa: BLE001
            pass
        scoring = _passthrough_scoring
    if isinstance(scoring, base_class):
        return scoring
    if callable(scoring):
        # We wrap the scorer, unless the user already supplied an instance of the Scorer class (or subclass)
        return base_class(scoring)
    raise ValueError("A valid scorer must either be a instance of `Scorer` (or subclass), None, or a callable.")


_non_homogeneous_scoring_error = ValidationError(
    "The returned score values are not homogeneous. "
    "For some datapoint a single values was returned, but for others a dictionary was returned. "
    "This is not allowed."
)


def _check_and_invert_score_dict(
    #  I don't care that this is to complex, some things need to be complex
    scores: list[ScoreTypeT[T]],
    default_agg: type[Aggregator],
) -> Union[tuple[type[Aggregator[T]], list[T]], dict[str, tuple[type[Aggregator[T]], list[T]]]]:
    """Invert the scores dictionary to a list of scores."""
    first_score = scores[0]
    if not isinstance(first_score, dict):
        # We expect a single score value from each datapoint
        # We check that no other value is a dictionary.
        # Other than that we can check nothing else here.
        # What datatypes are really allowed is controlled by the aggregator and will be checked there.
        if any(isinstance(s, dict) for s in scores):
            raise _non_homogeneous_scoring_error
        # We check that the score types are consistent for all datatypes.
        types = cast(set[type[Aggregator]], {(type(s) if isinstance(s, Aggregator) else default_agg) for s in scores})
        if len(types) > 1:
            raise ValidationError(
                f"The score values are not consistent. The following aggregation types were found: {types}"
            )
        consistent_type = types.pop()
        return (
            consistent_type,
            [(s.get_value() if isinstance(s, consistent_type) else consistent_type(s).get_value()) for s in scores],
        )

    score_types = {k: (type(s) if isinstance(s, Aggregator) else default_agg) for k, s in first_score.items()}
    return_dict = {k: (v, []) for k, v in score_types.items()}
    for s in scores:
        if not isinstance(s, dict):
            raise _non_homogeneous_scoring_error
        # We check that the score types are consistent for all datatypes.
        for k, v in score_types.items():
            try:
                score = s[k]
            except KeyError as e:
                raise ValidationError(
                    "The return values of the scoring function is expected to have the same keys for each datapoint. "
                    f"However, for at least one datapoint the return values is missing the key '{k}'. "
                    f"Expected keys are: {tuple(score_types.keys())}"
                ) from e
            # If the score is not wrapped in an aggregator, we wrap it in the default aggregator.
            if not isinstance(score, Aggregator):
                # NOTE: Not really elegant to wrap the values and then get the value two lines below... But it works.
                score = default_agg(score)
            if not isinstance(score, v):
                raise ValidationError(
                    f"The score value for {k} is not consistently of type {v}."
                    f"For at least one datapoint, the score value was {s[k]}."
                )
            return_dict[k][1].append(score.get_value())

    return return_dict



"""Helper for caching related tasks."""
import binascii
import functools
import multiprocessing
import warnings
from collections.abc import Sequence
from typing import Callable, Generic, Optional, TypeVar, Union

from joblib import Memory

from tpcp import Algorithm, get_action_methods_names, get_results, make_action_safe
from tpcp._hash import custom_hash

if multiprocessing.parent_process() is None:
    # We want to avoid spamming the user with warnings if they are running multiple processes
    warnings.warn(
        "Global caching is a little tricky to get right and our implementation is not yet battle-tested. "
        "Please double check that the results are correct and report any issues you find.",
        UserWarning,
        stacklevel=2,
    )

_instance_level_disk_cache_key = "__tpcp_disk_cached_action_method"
_class_level_lru_cache_key = "__tpcp_lru_cached_action_method"

T = TypeVar("T")


def _get_primary_action_method(algorithm_object: type[Algorithm]):
    action_names = get_action_methods_names(algorithm_object)
    if len(action_names) > 1:
        raise NotImplementedError("Caching is only implemented for algorithms with a single action method.")
    primary_action_name = action_names[0]
    return primary_action_name, getattr(algorithm_object, primary_action_name)


class UniversalHashableWrapper(Generic[T]):
    """A wrapper that makes any object hashable.

    The primary use case is to make function arguments hashable so that they can be used with `functools.lru_cache`.
    Under the hood we use the same pickle based hashing developed for joblib and extended for tpcp to handle torch and
    tensorflow objects.


    Examples
    --------
    >>> import functools
    >>> from tpcp.caching import UniversalHashableWrapper
    >>> import pandas as pd
    >>>
    >>> @functools.lru_cache(maxsize=1)
    ... def add(a: UniversalHashableWrapper[pd.DataFrame], b: UniversalHashableWrapper[pd.DataFrame]):
    ...     return a.obj + b.obj
    >>> df1 = pd.DataFrame({"a": [1, 2, 3]})
    >>> df2 = pd.DataFrame({"a": [4, 5, 6]})
    >>> add(UniversalHashableWrapper(df1), UniversalHashableWrapper(df2))

    """

    def __init__(self, obj: T) -> None:
        self.obj = obj

    def __hash__(self):
        """Hash the object using the pickle based approach."""
        return int(binascii.hexlify(custom_hash(self.obj).encode("utf-8")), 16)

    def __eq__(self, other):
        """Compare the object using their hash."""
        return custom_hash(self.obj) == custom_hash(other.obj)


def global_disk_cache(memory: Memory = Memory(None), *, cache_only: Optional[Sequence[str]] = None):
    """Wrap an algorithm/pipeline class to enable joblib based disk cashing for its primary action method.

    This will replace the action of the algorithm with a cached version.
    We use our knowledge about what is considered a "parameter" of an algorithm to clearly define the inputs of this
    cached function and avoid potential errors that might occur in a general caching solution.

    The cache will be invalidated, if any of the algorithm parameters change.

    This function can be used as a decorator on the class directly or called with the class as argument.
    In both cases the single global class object will be modified.
    At the moment there is no way of disabling the caching again.

    .. warning:: When using this decorator, all actions calls are not made on the original object, but on a clone, and
       only the results are "re-attached" to the original object.
       In case you rely on side effects of the action method, this might cause issues.
       But, if you are relying on side effects, you are probably doing something wrong anyway.

    Parameters
    ----------
    memory
        The joblib memory object that is used to cache the results.
        ``Memory(None)`` is equivalent to no caching.
    cache_only
        A list of strings that defines which results should be cached.
        If None, all results are cached.
        In case you only cash a subset of the results, only these results will be available on the returned objects.
        Also, during the first (uncached) call, only the results that are cached will be available.
        This might be helpful to reduce the size of the cache.

    Returns
    -------
    The algorithm class with the cached action method.

    See Also
    --------
    tpcp.caching.global_ram_cache
        Same as this function, but uses an LRU cache in RAM instead of a disk cache.
    """

    def inner(algorithm_object: type[Algorithm]):
        # This only return the first action method, but this is fine for now
        # This method is "unbound", as we are working on the class, not an instance
        primary_action_name, action_method = _get_primary_action_method(algorithm_object)

        # We need to make the action method safe, as we are going to do weird stuff that expects correct implementation
        action_method = make_action_safe(action_method)

        @functools.wraps(action_method)
        def wrapper(self, *args_outer, **kwargs_outer):
            if getattr(self, _instance_level_disk_cache_key, None) is None:
                # our cached function gets the args, the kwargs and "fake self" as input.
                # Fake self is a clean clone of the algorithm instance.
                # It basically only encodes the parameters and the "name" of the algorithm.
                def cachable_inner(__fake_self, __cache_only, *args, **kwargs):
                    after_action_instance: Algorithm = action_method(__fake_self, *args, **kwargs)
                    # We return the results instead of the instance, as the results can be easily pickled.
                    if __cache_only is None:
                        return get_results(after_action_instance)
                    return {k: v for k, v in get_results(after_action_instance).items() if k in __cache_only}

                cached_inner = memory.cache(cachable_inner)
                setattr(self, _instance_level_disk_cache_key, cached_inner)

            cached_inner = getattr(self, _instance_level_disk_cache_key)
            results = cached_inner(self.clone(), cache_only, *args_outer, **kwargs_outer)

            # manually "glue" the results back to the instance
            for result_name, result in results.items():
                setattr(self, result_name, result)

            return self

        wrapper.__wrapped__ = action_method

        setattr(algorithm_object, primary_action_name, wrapper)
        return algorithm_object

    return inner


def remove_disk_cache(algorithm_object: type[Algorithm]):
    """Remove the disk cache from an algorithm class."""
    primary_action_name, action_method = _get_primary_action_method(algorithm_object)

    if getattr(action_method, "__wrapped__", None) is not None:
        setattr(algorithm_object, primary_action_name, action_method.__wrapped__)

    return algorithm_object


def global_ram_cache(max_n: Optional[int] = None, *, cache_only: Optional[Sequence[str]] = None):
    """Wrap an algorithm/pipeline class to enable LRU based RAM cashing for its primary action method.

    .. warning:: When using this decorator, all actions calls are not made on the original object, but on a clone, and
       only the results are "re-attached" to the original object.
       In case you rely on side effects of the action method, this might cause issues.
       But, if you are relying on side effects, you are probably doing something wrong anyway.

    Parameters
    ----------
    max_n
        The maximum number of entries in the cache.
        If None, the cache will grow without limit.
    cache_only
        A list of strings that defines which results should be cached.
        If None, all results are cached.
        In case you only cash a subset of the results, only these results will be available on the returned objects.
        Also, during the first (uncached) call, only the results that are cached will be available.
        This might be helpful to reduce the size of the cache.

    Returns
    -------
    The algorithm class with the cached action method.

    See Also
    --------
    tpcp.caching.global_disk_cache
        Same as this function, but uses a disk cache instead of an LRU cache in RAM.

    """
    if cache_only is not None:
        cache_only = tuple(cache_only)

    def inner(algorithm_object: type[Algorithm]):
        primary_action_name, action_method = _get_primary_action_method(algorithm_object)

        # We need to make the action method safe, as we are going to do weird stuff that expects correct implementation
        action_method = make_action_safe(action_method)

        @functools.wraps(action_method)
        def wrapper(self, *args_outer, **kwargs_outer):
            if getattr(self, _class_level_lru_cache_key, None) is None:

                def cachable_inner(__fake_self, __cache_only, hashable_args, hashable_kwargs):
                    after_action_instance: Algorithm = action_method(
                        # Note: We need to clone the object here again.
                        # the LRU cache seems to calculate the hash of the object only after the function call.
                        # So we need to make sure that the object is not modified in the meantime.
                        __fake_self.obj.clone(),
                        *hashable_args.obj,
                        **hashable_kwargs.obj,
                    )
                    if __cache_only is None:
                        return get_results(after_action_instance)
                    return {k: v for k, v in get_results(after_action_instance).items() if k in __cache_only}

                # Create a LRUCache with max_n entries and store it on the class.
                # This way it is unique and shared between all instances of the class.
                cached_inner = {"cached_func": functools.lru_cache(max_n)(cachable_inner)}
                setattr(algorithm_object, _class_level_lru_cache_key, cached_inner)

            cached_inner = getattr(self, _class_level_lru_cache_key)

            results = cached_inner["cached_func"](
                UniversalHashableWrapper(self.clone()),
                cache_only,
                UniversalHashableWrapper(args_outer),
                UniversalHashableWrapper(kwargs_outer),
            )

            # manually "glue" the results back to the instance
            for result_name, result in results.items():
                setattr(self, result_name, result)

            return self

        wrapper.__wrapped__ = action_method

        setattr(algorithm_object, primary_action_name, wrapper)
        return algorithm_object

    return inner


def remove_ram_cache(algorithm_object: type[Algorithm]):
    """Remove the RAM cache from an algorithm class."""
    primary_action_name, action_method = _get_primary_action_method(algorithm_object)

    if getattr(action_method, "__wrapped__", None) is not None:
        setattr(algorithm_object, primary_action_name, action_method.__wrapped__)

    if getattr(algorithm_object, _class_level_lru_cache_key, None) is not None:
        delattr(algorithm_object, _class_level_lru_cache_key)

    return algorithm_object


def remove_any_cache(algorithm_object: type[Algorithm]):
    """Remove any cache from an algorithm class."""
    return remove_disk_cache(remove_ram_cache(algorithm_object))


def get_ram_cache_obj(algorithm_object: type[Algorithm]):
    """Get the RAM cache object from an algorithm class."""
    return getattr(algorithm_object, _class_level_lru_cache_key, None)["cached_func"]


_GLOBAL_CACHE_REGISTRY: dict[tuple[str, str], Callable] = {}


def hybrid_cache(
    joblib_memory: Memory = Memory(None),
    lru_cache_maxsize: Union[Optional[int], bool] = False,
):
    """Cache a function using joblib memory and a lru cache at the same time.

    This function attempts to be the best of both worlds and uses joblib.Memory to cache function calls between runs
    and a lru_cache to cache function calls during a run.

    When the cached function is called, the lookup will work as follows:

    1. Is the function result in the lrucache? If yes, return it.
    2. Is the function result in the joblib memory? If yes, return it and cache it in the lru cache.
    3. Call the function and cache it in the joblib memory and the lru cache. Return the result.

    It further solves one of the issues that you might run into with ``lru_cache``, that it is difficult to create a
    wrapped function during runtime, as calling ``lru_cache`` directly will create a new cache for each call.
    We work around this by using a global cache that stores the wrapped functions.
    The cache key is a tuple of the function name and a hash of the function, the joblib memory and the lru_cache paras.
    This means, if you create a new cache with different cache parameters, you will get a new cache, but if you call
    ``staggered_cache`` with the same parameters, you will get the same object back.

    You can access this global cache via the ``__cache_registry__`` attribute of this function
    (``staggered_cache.__cache_registry__``).

    Parameters
    ----------
    joblib_memory
        The joblib memory object that is used to cache the results.
        ``Memory(None)`` is equivalent to no caching.
    lru_cache_maxsize
        The maximum number of entries in the cache.
        If None, the cache will grow without limit.
        If False, no lru_cache is used.

    Returns
    -------
    caching_decorator
        A decorator that can be used to cache a function with the given parameters.

    Examples
    --------
    >>> import pandas as pd
    >>> from tpcp.caching import hybrid_cache
    >>> from joblib import Memory
    >>>
    >>> @hybrid_cache(Memory(".cache"), lru_cache_maxsize=1)
    ... def add(a: pd.DataFrame, b: pd.DataFrame):
    ...     return a + b
    >>> df1 = pd.DataFrame({"a": [1, 2, 3]})
    >>> df2 = pd.DataFrame({"a": [4, 5, 6]})
    >>> add(df1, df2)

    """

    def inner(function: Callable):
        paras_hash = custom_hash((function.__name__, id(function), joblib_memory, lru_cache_maxsize))
        cache_key = (function.__name__, paras_hash)
        if cache_key in _GLOBAL_CACHE_REGISTRY:
            return _GLOBAL_CACHE_REGISTRY[cache_key]

        if lru_cache_maxsize is False:

            @functools.wraps(function)
            def final_wrapped(*args, **kwargs):
                return joblib_memory.cache(function)(*args, **kwargs)
        else:

            def inner_cached(*hash_safe_args, **hash_safe_kwargs):
                args = tuple(arg.obj for arg in hash_safe_args)
                kwargs = {k: v.obj for k, v in hash_safe_kwargs.items()}
                return joblib_memory.cache(function)(*args, **kwargs)

            final_cached = functools.lru_cache(lru_cache_maxsize)(inner_cached)

            @functools.wraps(function)
            def final_wrapped(*args, **kwargs):
                hash_safe_args = tuple(UniversalHashableWrapper(arg) for arg in args)
                hash_safe_kwargs = {k: UniversalHashableWrapper(v) for k, v in kwargs.items()}
                return final_cached(*hash_safe_args, **hash_safe_kwargs)

        _GLOBAL_CACHE_REGISTRY[cache_key] = final_wrapped

        return final_wrapped

    return inner


hybrid_cache.__cache_registry__ = _GLOBAL_CACHE_REGISTRY


__all__ = [
    "global_disk_cache",
    "global_ram_cache",
    "UniversalHashableWrapper",
    "remove_disk_cache",
    "remove_ram_cache",
    "remove_any_cache",
    "get_ram_cache_obj",
    "hybrid_cache",
]



"""A mixin for all common tests that should be run on all algorithm classes."""
import inspect
from typing import Generic

import joblib
import pytest
from numpydoc.docscrape import NumpyDocString

from tpcp import (
    Algorithm,
    get_action_method,
    get_action_params,
    get_param_names,
    get_results,
    make_action_safe,
)
from tpcp._base import BaseTpcpObjectT
from tpcp._hash import custom_hash


class TestAlgorithmMixin(Generic[BaseTpcpObjectT]):
    """A mixin for all common tests that should be run on all algorithm classes.

    You can use this mixin to test your algorithm class OR Pipeline sub-class (as pipelines are just algorithms).

    This mixin is intended to be used in a pytest testclass.
    For this, create a new class that inherits from this mixin and set the ALGORITHM_CLASS attribute to the algorithm
    class that should be tested.
    Then set `__test__ = True` to enable the tests.

    You can further customize the tests by setting the following attributes:

    - ONLY_DEFAULT_PARAMS: If set to False, the testclass can have non-optional kwargs in the constructor.
    - ASSUME_HASHABLE: If set to False, the testclass is not expected to be hashable by joblib.
    - CHECK_DOCSTRING: If set to False, the docstring tests are skipped.

    In some very specific cases you might want to ignore some parameters in the docstring tests.
    For this, set the _IGNORED_NAMES attribute to a tuple of parameter names that should be ignored.

    Examples
    --------
    >>> class TestMyAlgorithm(TestAlgorithmMixin):
    ...     ALGORITHM_CLASS = MyAlgorithm
    ...     __test__ = True
    ...
    ...     ONLY_DEFAULT_PARAMS = False

    """

    ALGORITHM_CLASS: type[BaseTpcpObjectT]
    __test__ = False

    ONLY_DEFAULT_PARAMS: bool = True
    ASSUME_HASHABLE: bool = True
    CHECK_DOCSTRING: bool = True

    _IGNORED_NAMES = ()

    @pytest.fixture()
    def after_action_instance(self, **kwargs) -> BaseTpcpObjectT:  # noqa: PT004
        """Return an instance of the algorithm class AFTER the action is performed.

        This needs to be implemented by every testclass.
        The returned algorithm instance should have the result attributes and the "other parameters" (i.e. the action
        method inputs) set.
        """
        raise NotImplementedError

    def test_is_algorithm(self):
        """Test that the class is actually an algorithm."""
        assert issubclass(self.ALGORITHM_CLASS, Algorithm)

    def test_after_action_instance_valid(self, after_action_instance):
        """Test that the implemented after_action_fixture returns a valid instance."""
        assert isinstance(after_action_instance, Algorithm)
        assert isinstance(after_action_instance, self.ALGORITHM_CLASS)

    def test_init(self):
        """Test that all init paras are passed through untouched."""
        field_names = get_param_names(self.ALGORITHM_CLASS)
        init_dict = {k: k for k in field_names}

        test_instance = self.ALGORITHM_CLASS(**init_dict)

        for k, v in init_dict.items():
            assert getattr(test_instance, k) == v, k

    def test_only_optional_kwargs(self):
        """Test that the class has only optional kwargs."""
        if self.ONLY_DEFAULT_PARAMS is False:
            pytest.skip("The testclass is not expected to have only optional kwargs.")
        self.ALGORITHM_CLASS()

    def test_algorithm_can_be_cloned(self, after_action_instance):
        """Test if an algorithm can be cloned."""
        initial = after_action_instance.clone()
        assert custom_hash(initial) == custom_hash(initial.clone())

    def test_all_parameters_documented(self):
        """Test if all init-parameters are documented."""
        if not self.CHECK_DOCSTRING:
            pytest.skip("Docstring testing explicitly disabled.")
        docs = NumpyDocString(inspect.getdoc(self.ALGORITHM_CLASS))

        documented_names = {p.name for p in docs["Parameters"]}
        documented_names -= set(self._IGNORED_NAMES)
        actual_names = set(get_param_names(self.ALGORITHM_CLASS))
        actual_names -= set(self._IGNORED_NAMES)

        assert documented_names == actual_names

    def test_all_attributes_documented(self, after_action_instance):
        """Test if all result attributes are documented."""
        if not self.CHECK_DOCSTRING:
            pytest.skip("Docstring testing explicitly disabled.")
        if not after_action_instance:
            pytest.skip("The testclass did not implement the correct `after_action_instance` fixture.")
        docs = NumpyDocString(inspect.getdoc(self.ALGORITHM_CLASS))

        documented_names = {p.name for p in docs["Attributes"]}
        documented_names -= set(self._IGNORED_NAMES)
        actual_names = set(get_results(after_action_instance).keys())
        actual_names -= set(self._IGNORED_NAMES)

        assert documented_names == actual_names

    def test_all_other_parameters_documented(self, after_action_instance):
        """Test if all other parameters (action method inputs) are documented."""
        if not self.CHECK_DOCSTRING:
            pytest.skip("Docstring testing explicitly disabled.")
        if not after_action_instance:
            pytest.skip("The testclass did not implement the correct `after_action_instance` fixture.")
        docs = NumpyDocString(inspect.getdoc(self.ALGORITHM_CLASS))

        documented_names = {p.name for p in docs["Other Parameters"]}
        documented_names -= set(self._IGNORED_NAMES)
        actual_names = set(get_action_params(after_action_instance).keys())
        actual_names -= set(self._IGNORED_NAMES)

        assert documented_names == actual_names

    def test_action_method_returns_self(self, after_action_instance):
        """Test if the action method returns self."""
        # call the action method a second time to test the output
        parameters = get_action_params(after_action_instance)
        results = get_action_method(after_action_instance)(**parameters)

        assert id(results) == id(after_action_instance)

        assert id(results) == id(after_action_instance)

    def test_set_params_valid(self, after_action_instance):
        """Test that set_params works."""
        instance = after_action_instance.clone()
        valid_names = get_param_names(instance)
        values = list(range(len(valid_names)))
        instance.set_params(**dict(zip(valid_names, values)))

        for k, v in zip(valid_names, values):
            assert getattr(instance, k) == v, k

    def test_set_params_invalid(self, after_action_instance):
        """Test that set_params raises an error if an invalid parameter is passed."""
        instance = after_action_instance.clone()

        with pytest.raises(ValueError, match="`an_invalid_name` is not a valid parameter") as e:
            instance.set_params(an_invalid_name=1)

        assert "an_invalid_name" in str(e)
        assert self.ALGORITHM_CLASS.__name__ in str(e)

    def test_hashing(self, after_action_instance):
        """Check if caching with joblib will work as expected."""
        if self.ASSUME_HASHABLE is False:
            pytest.skip("The testclass is not expected to be hashable.")
        instance = after_action_instance.clone()

        assert joblib.hash(instance) == joblib.hash(instance.clone())

    def test_passes_safe_action_checks(self, after_action_instance):
        """Test that the algorithm passes the safe action checks."""
        # We just wrap the method and call it.
        # We don't care about the return value
        make_action_safe(get_action_method(after_action_instance))(
            after_action_instance,
            **get_action_params(after_action_instance),
        )



"""Baseclass for optimizers.

This is in a separate file to avoid circular imports.
"""
from typing import Any, ClassVar, Generic, Union

from typing_extensions import Self

from tpcp import Algorithm, Parameter
from tpcp._dataset import DatasetT
from tpcp._pipeline import PipelineT


class BaseOptimize(Algorithm, Generic[PipelineT, DatasetT]):
    """Base class for all optimizer."""

    _action_methods: ClassVar[str] = "optimize"

    pipeline: Parameter[PipelineT]

    dataset: DatasetT

    optimized_pipeline_: PipelineT

    def optimize(self, dataset: DatasetT, **optimize_params: Any) -> Self:
        """Apply some form of optimization on the input parameters of the pipeline."""
        raise NotImplementedError()

    def run(self, datapoint: DatasetT) -> PipelineT:
        """Run the optimized pipeline.

        This is a wrapper to contain API compatibility with `Pipeline`.
        """
        return self.optimized_pipeline_.run(datapoint)

    def safe_run(self, datapoint: DatasetT) -> PipelineT:
        """Run the optimized pipeline.

        This is a wrapper to contain API compatibility with `Pipeline`.
        """
        return self.optimized_pipeline_.safe_run(datapoint)

    def score(self, datapoint: DatasetT) -> Union[float, dict[str, float]]:
        """Run score of the optimized pipeline.

        This is a wrapper to contain API compatibility with `Pipeline`.
        """
        return self.optimized_pipeline_.score(datapoint)



"""Private base classes for tpcp.

These classes are in a separate module to avoid circular imports.
In basically all cases, you do not need them.
"""
from __future__ import annotations

import contextlib
import copy
import dataclasses
import inspect
import os
import sys
import warnings
from collections import defaultdict
from functools import wraps
from pathlib import Path
from types import MethodWrapperType
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Callable,
    ClassVar,
    Generic,
    Optional,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from typing_extensions import (
    Concatenate,
    Literal,
    ParamSpec,
    Self,
    get_args,
    get_origin,
)

from tpcp._parameters import _ParaTypes
from tpcp.exceptions import (
    MutableDefaultsError,
    PotentialUserErrorWarning,
    ValidationError,
)

try:
    import tensorflow as tf
except ImportError:
    tf = None

if TYPE_CHECKING:
    from collections.abc import Iterable

T = TypeVar("T")
P = ParamSpec("P")
BaseTpcpObjectT = TypeVar("BaseTpcpObjectT", bound="BaseTpcpObject")
_BaseTpcpObjectT = TypeVar("_BaseTpcpObjectT", bound="_BaseTpcpObject")


class _Nothing:
    """Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    ``_Nothing`` is a singleton. There is only ever one of it.

    This implementation is taken from the attrs package.
    """

    _singleton: Optional[_Nothing] = None

    def __new__(cls) -> _Nothing:
        if _Nothing._singleton is None:
            _Nothing._singleton = super().__new__(cls)
        return _Nothing._singleton

    def __repr__(self) -> str:
        return "NOTHING"

    def __bool__(self) -> Literal[False]:
        return False

    def __len__(self) -> Literal[0]:
        return 0  # __bool__ for Python 2


NOTHING = _Nothing()
"""
Sentinel to indicate the lack of a value when ``None`` is ambiguous.
"""


def _get_init_defaults(cls: type[_BaseTpcpObject]) -> dict[str, inspect.Parameter]:
    # fetch the constructor or the original constructor before deprecation wrapping if any
    init = cls.__init__
    if init is object.__init__ or isinstance(init, MethodWrapperType):
        # No explicit constructor to introspect
        return {}

    # introspect the constructor arguments to find the model parameters to represent
    init_signature = inspect.signature(init)
    # Consider the constructor parameters excluding 'self'
    defaults = {k: p for k, p in init_signature.parameters.items() if p.name != "self" and p.kind != p.VAR_KEYWORD}
    return defaults


def _replace_defaults_wrapper(
    cls: type[_BaseTpcpObjectT], old_init: Callable[Concatenate[_BaseTpcpObjectT, P], T]
) -> Callable[Concatenate[_BaseTpcpObjectT, P], T]:
    """Decorate an init to create new instances of mutable defaults.

    This should only be used in combination with `default` and will be applied as part of `__init_subclass`.
    Direct usage of this decorator should not be required.
    """
    # We get the params of the old inits here and not at runtime to avoid issues when the class is subclassed and
    # super().__init__ is called before all parameters of the child object are set.
    # This way, the param checks only concern the parameters of the current class.
    params = get_param_names(cls)

    @wraps(old_init)
    def new_init(self: _BaseTpcpObjectT, *args: P.args, **kwargs: P.kwargs) -> None:
        # call the old init.
        old_init(self, *args, **kwargs)

        # Check if any of the initial values has a "default parameter flag".
        # If yes we replace it with a clone (in case of a tpcp object) or a deepcopy in case of other objects.
        # This is handled by the factory `get_value` method.
        for p in params:
            if isinstance(val := getattr(self, p), BaseFactory):
                setattr(self, p, val.get_value())

    # This is just for introspection, in case we want to know if we have a modified init.
    new_init.__tpcp_wrapped__ = True

    return cast(Callable[Concatenate[_BaseTpcpObjectT, P], T], new_init)


def _retry_eval_with_missing_locals(
    expression: str,
    globalns: Optional[dict[str, Any]] = None,
    localns: Optional[dict[str, Any]] = None,
) -> Any:
    globalns = globalns or {}
    localns = localns or {}
    # We make a copy to not overwrite the input dict
    localns = {**localns}

    # That seems scary :D Let's see if this causes any issues
    # We use a value here instead of a "while True" to not get the program stuck in an endless loop.
    for _ in range(100):
        try:
            val = eval(expression, globalns, localns)  # noqa: PGH001
            break
        except NameError as e:
            missing = str(e).split("'")[1]
            localns[missing] = None
        except AttributeError as e:
            raise RuntimeError(
                "You ran into an edge case of the builtin type resolver. "
                "This happens if you use a nested type annotation that is only valid during runtime. "
                "This usually happens if you are using a `if TYPE_CHECKING:` guard for some of your imports to avoid "
                "circular dependencies.\n"
                "For most of these cases we have a built in workaround, but only if you provide the type directly and "
                "not in an attribute notation:\n"
                "\n"
                "This is not allowed:"
                "\n"
                ">>> if TYPE_CHECKING:\n"
                "...     import my\n"
                ">>> MyClass:\n"
                "...     para: my.custom_type\n"
                "\n\n"
                "This should work:\n"
                ">>> if TYPE_CHECKING:\n"
                "...     from my import custom_type\n"
                ">>> MyClass:\n"
                "...     para: custom_type\n"
                "\n\n"
                "And this as well:\n"
                ">>> if TYPE_CHECKING:\n"
                "...     import my\n"
                "...     custom_type_var = my.custom_type\n"
                ">>> MyClass:\n"
                "...     para: custom_type_var\n"
            ) from e
    else:
        raise RuntimeError(
            "Trying to resolve Parameter Type hints has resulted in an unexpected issue. "
            f"We were trying to evaluate the expression: {expression} . "
            "It seems like this expression contained forward references that could not be resolved. "
            "This should not happen!"
            "Please open an issue on GitHub with an code example that results in this issue. "
            "Sry for the inconvenience :)"
        )
    return val


def _custom_get_type_hints(cls: type[_BaseTpcpObject]) -> dict[str, Any]:
    """Extract type hints while avoiding issues with forward references.

    We automatically skip all douple-underscore methods.
    """
    hints = {}
    for base in reversed(cls.__mro__):
        base_globals = sys.modules[base.__module__].__dict__
        ann = base.__dict__.get("__annotations__", {})
        for name, value in ann.items():
            if name.startswith("__"):
                continue
            if value is None:
                value = type(None)  # noqa: PLW2901
            elif isinstance(value, str):
                # NOTE: This does not check if the str is a valid expression.
                #       This might not be an issue, but could lead to obscure error messages.
                value = _retry_eval_with_missing_locals(value, base_globals)  # noqa: PLW2901
            hints[name] = value
    return hints


def _extract_annotations(
    cls: type[_BaseTpcpObject], init_fields: dict[str, inspect.Parameter]
) -> dict[str, _ParaTypes]:
    cls_annotations = _custom_get_type_hints(cls)
    para_annotations = {}
    for k, v in cls_annotations.items():
        origin = get_origin(v)
        if origin is ClassVar:
            # If the parameter is a ClassVar, we go one level deeper and check, if its argument was annotated.
            v = get_args(v)[0]  # noqa: PLW2901
            origin = get_origin(v)
        if origin is Annotated:
            for annot in get_args(v)[1:]:
                if isinstance(annot, _ParaTypes):
                    para_annotations[k] = annot
                    break
        elif k in init_fields:
            para_annotations[k] = _ParaTypes.SIMPLE
    return para_annotations


def _validate_parameter(cls: type[_BaseTpcpObject]):
    # We extract the fields of the init
    fields = _get_init_defaults(cls)

    # Validation
    _has_dangerous_mutable_default(fields, cls)
    _has_invalid_name(fields, cls)

    if (
        dataclasses.is_dataclass(cls)
        or getattr(
            cls, "__attrs_attrs__", False
        )  # This checks if the class it an `attrs` class, without importing `attrs`
    ) and any(isinstance(field.default, BaseFactory) for field in fields.values()):
        # When you are already using dataclasses or attrs, you must use their factory methods, as we don't overwrite
        # the init to support CloneFactory anymore.
        raise ValidationError(
            "You are using the tpcp default factory (`cf`/`CloneFactory`) in combination with "
            "`dataclasses` or `attrs`. "
            "Use the `default_factory`/`factory` option of `dataclasses.field`/`attrs.field` instead."
        )


def _validate_all_parent_parameters_implemented(cls: type[_BaseTpcpObject]):
    """Validate that a class implements all parameters of its parents.

    We don't raise an error in this case, we just warn the user.
    """
    param_names = set(get_param_names(cls))

    for base in cls.__bases__:
        if not issubclass(base, _BaseTpcpObject):
            continue
        base_params = set(get_param_names(base))
        if not base_params.issubset(param_names):
            missing_params = base_params - param_names
            warnings.warn(
                f"The class {cls.__name__} does not implement all parameters of its parent {base.__name__}. "
                f"Missing parameters: {missing_params}\n"
                "This might not be a problem, but indicates bad design and you might run into actual issues with some "
                "of the validation magic `tpcp` does in the background. "
                "We would recommend to implement all parameters of your parents in a subclass.",
                stacklevel=2,
            )


def _get_tpcp_validated(cls_or_instance: Union[type[_BaseTpcpObject], _BaseTpcpObject]):
    cls = cls_or_instance if isinstance(cls_or_instance, type) else type(cls_or_instance)
    return cls.__dict__.get("__tpcp_validated_hidden__", False)


def _set_tpcp_validated(cls_or_instance: Union[type[_BaseTpcpObject], _BaseTpcpObject], value: bool):
    cls = cls_or_instance if isinstance(cls_or_instance, type) else type(cls_or_instance)
    cls.__tpcp_validated_hidden__ = value


class _BaseTpcpObject:
    # These two parameters should be initialized once per class.
    # This means, when we read them, we always check in the dictionary if they are defined.
    __field_annotations_cache__: dict[str, _ParaTypes]
    __tpcp_validated_hidden__: bool

    @property
    def __field_annotations__(self) -> dict[str, _ParaTypes]:
        """Get the field annotations that provide higher level information about the role of a parameter.

        We implement that as property to move the extraction of the annotations as far down in the initialization as
        possible, when we are sure the cls definition is properly finalized (i.e. dataclass decorators are run)
        """
        cls = type(self)
        # We check if we have a cache, and if it is our cache or the cache inherited by the parent class.
        # If it is from the parent class, we make sure, we create our own.
        if cache := cls.__dict__.get("__field_annotations_cache__", None):
            return cache
        # We need to create a new cache!
        fields = _get_init_defaults(cls)
        field_annotations = _extract_annotations(cls, fields)
        # We only validate the annotations here. If the user is not using the annotations, no need to bother them
        # with useless error messages
        _annotations_are_valid(field_annotations, fields, cls_name=cls.__name__)
        # We set the cache on the cls, so that only one instance is required to process it.
        cls.__field_annotations_cache__ = field_annotations
        return field_annotations

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        # If the class has no init or did inherit its init from the parent (i.e. it is not in its own dict), there is
        # nothing to do.
        if "__init__" not in cls.__dict__:
            return

        # If we have a custom init, we need to wrap it with a method that replaces our default values on init.
        setattr(cls, "__init__", _replace_defaults_wrapper(cls, cls.__init__))

    @classmethod
    def __custom_tpcp_validation__(cls):
        """Trigger custom validation.

        This method can be overwritten by the user to implement custom validation.
        The validation is triggered on the first call of `get_params.
        Within the method, either raise a `ValidationError` or privide warnings to the user.
        """
        return


class BaseTpcpObject(_BaseTpcpObject):
    """Baseclass for all tpcp objects."""

    _composite_params: ClassVar[tuple[str, ...]] = ()

    def get_params(self, deep: bool = True) -> dict[str, Any]:
        """Get parameters for this algorithm.

        Parameters
        ----------
        deep
            Only relevant if object contains nested algorithm objects.
            If this is the case and deep is True, the params of these nested objects are included in the output using a
            prefix like `nested_object_name__` (Note the two "_" at the end)

        Returns
        -------
        params
            Parameter names mapped to their values.

        """
        return _get_params(self, deep)

    def set_params(self, **params: Any) -> Self:
        """Set the parameters of this Algorithm.

        To set parameters of nested objects use `nested_object_name__para_name=`.
        """
        return _set_params(self, **params)

    def clone(self) -> Self:
        """Create a new instance of the class with all parameters copied over.

        This will create a new instance of the class itself and all nested objects
        """
        return clone(self, safe=True)

    def __repr_parameter__(self, name: str, value: Any) -> str:
        """Format a parameter for the __repr__ method.

        This method is used to format the parameters for the __repr__ method.
        It is called for each parameter and can be overwritten to customize the output.

        Parameters
        ----------
        name
            The name of the parameter
        value
            The value of the parameter

        Returns
        -------
        str
            The formatted string for the parameter

        """
        return f"{name}={value!r}"

    def __repr__(self) -> str:
        """Provide generic representation for the object based on all parameters."""
        class_name = type(self).__name__
        paras = self.get_params(deep=False)
        result = [class_name, "("]
        first = True
        for name, para in paras.items():
            if first:
                first = False
            else:
                result.append(", ")
            result.append(self.__repr_parameter__(name, para))
        return "".join(result) + ")"

    @classmethod
    def __clone_param__(cls, param_name: str, value: Any) -> Any:  # pylint: disable=unused-argument
        """Handle cloning of specific parameters of the object.

        This method exists only to allow object the handling of "strange" python datatypes that can not be deepcopied
        for some reason.
        This can happen when library authors overwrite deepcopy behaviour or if handling objects can not be easily
        deepcopied as there are implemented in Cython (or similar).

        In such cases you can overwrite how certain objects should be cloned using this method.
        Note, that you are on your own, when you do that.
        I.e. we will not check that you do it correctly and that you ensure that the new object is really independent of
        the old or any other things you might run into.

        If you are using this method, we recommend to at least check that the cloned object is identical to the
        original based on our implemented hashing method.

        >>> from tpcp._hash import custom_hash
        >>> assert custom_hash(my_obj) == custom_hash(my_obj.clone())

        If this doesn't pass, many other features in tpcp will not work as well.
        """
        return clone(value, safe=False)


def _get_deep_params(obj, parent_key) -> dict[str, Any]:
    # This is a little magic, that also gets the parameters of nested sklearn classifiers.
    if hasattr(obj, "get_params"):
        deep_items = obj.get_params(deep=True).items()
        return {parent_key + "__" + k: val for k, val in deep_items}
    return {}


def _assert_is_allowed_composite_value(val, parent_key: str, iteration: int):
    if (not isinstance(val, tuple)) or (len(val) != 2) or (not isinstance(val[0], str)):
        raise ValidationError(
            f"The provided parameters for the composite field {parent_key} does not seem to be the "
            "right type. "
            "It should be a sequence of `(name, value)` tuples, but the obj at position "
            f"{iteration} in the sequence was not a tuple but:\n`{val}`"
        )


def _recursive_validate(cls: type[_BaseTpcpObject]):
    if not _get_tpcp_validated(cls):
        _validate_parameter(cls)
        _validate_all_parent_parameters_implemented(cls)
        cls.__custom_tpcp_validation__()

        _set_tpcp_validated(cls, True)

        for base in cls.__bases__:
            if issubclass(base, _BaseTpcpObject):
                _recursive_validate(base)


def _get_params(instance: _BaseTpcpObject, deep: bool = True) -> dict[str, Any]:
    # At some point, we want to validate that the user defined the class correctly.
    # To allow for all strange modifications of the class, we run the validation, when we actually need the parameters.
    # This usually required a call to `get_params`.
    # Hence, we run the validation here.
    # To avoid running this validation on every get_params call, we use `__tpcp_validated__` to make sure it runs only
    # once per class.
    # We actually run this validation recusively for all base classes, in case `get_params` is never called on the base
    # directly.
    _recursive_validate(type(instance))

    valid_fields = get_param_names(type(instance))
    comp_fields = getattr(instance, "_composite_params", ())
    out: dict[str, Any] = {}
    for key in valid_fields:
        value = getattr(instance, key)
        if deep:
            if value is not None and comp_fields and key in comp_fields:
                # Here we handle composite field (i.e. fields that contain sequences of (name, tpcp_obj_instance)
                # tuples.
                # We basically flatten the list by storing the instance under the name `{key}__{name}` and then add
                # all the nested parameters
                for i, v in enumerate(value):
                    _assert_is_allowed_composite_value(v, key, i)
                    name, obj = v
                    nested_key = f"{key}__{name}"
                    out[nested_key] = obj
                    if deep:
                        out.update(**_get_deep_params(obj, nested_key))
            else:
                out.update(**_get_deep_params(value, key))
        out[key] = value
    return out


def _set_comp_field(instance, field_name, params):
    # We first partition our field names to know to which index they belong
    comp_params: defaultdict[str, Any] = defaultdict(dict)
    for key, value in params.items():
        key, delim, sub_key = key.partition("__")  # noqa: PLW2901
        if delim:
            comp_params[key][sub_key] = value
        else:
            comp_params[key]["*"] = value

    # Then we iterate over existing values in the compound field and recreate it
    new_list = []
    comp_list = list(getattr(instance, field_name))
    for key, old_value in comp_list:
        nested_values = comp_params.pop(key, {})
        new_value = old_value if "*" not in nested_values else nested_values.pop("*")
        if nested_values:
            new_value = new_value.set_params(**nested_values)
        new_list.append((key, new_value))
    if comp_params:
        # Some values are left over -> aka they are invalid
        raise ValueError(
            f"You are trying to set values for a compound field {field_name} with the identifiers "
            f"{list(comp_params.keys())}. "
            "These identifiers could not be found within the existing list of identifiers for this "
            f"field ({[c[0] for c in comp_list]}. "
            "We don't support setting params on not existent value in a compound field. "
            "Recreate the entire field if you want to add or delete entries from a compound field."
        )

    setattr(instance, field_name, new_list)


def _set_params(instance: BaseTpcpObjectT, **params: Any) -> BaseTpcpObjectT:
    """Set the parameters of an instance.

    To set parameters of nested objects use `nested_object_name__para_name=`.
    """
    # Basically copied from sklearn
    if not params:
        return instance
    valid_params = instance.get_params(deep=True)
    comp_fields = getattr(instance, "_composite_params", ())

    nested_params: defaultdict[str, Any] = defaultdict(dict)  # grouped by prefix
    for key, value in params.items():
        key, delim, sub_key = key.partition("__")  # noqa: PLW2901
        if key not in valid_params:
            raise ValueError(f"`{key}` is not a valid parameter name for {type(instance).__name__}.")

        if delim:
            nested_params[key][sub_key] = value
        else:
            setattr(instance, key, value)
            valid_params[key] = value

    for key, sub_params in nested_params.items():
        if key in comp_fields:
            _set_comp_field(instance, key, sub_params)
        else:
            valid_params[key].set_params(**sub_params)
    return instance


def get_param_names(obj: Union[type[_BaseTpcpObject], _BaseTpcpObject]) -> list[str]:
    """Get parameter names for the object.

    The parameters of an algorithm/pipeline are defined based on its `__init__` method.
    All parameters of this method are considered parameters of the algorithm.

    Notes
    -----
    Adopted based on :meth:`sklearn.base.BaseEstimator._get_param_names`.

    Returns
    -------
    param_names
        List of parameter names of the algorithm

    """
    cls = obj if isinstance(obj, type) else type(obj)
    parameters = list(_get_init_defaults(cls).values())
    for p in parameters:
        if p.kind == p.VAR_POSITIONAL:
            raise RuntimeError(
                "tpcp algorithms and pipelines should always specify their parameters in the signature of their "
                f"__init__ (no varargs). {cls} doesn't follow this convention."
            )
    # Extract and sort argument names excluding 'self'
    return sorted([p.name for p in parameters])


def _get_annotated_fields_of_type(
    instance_or_cls: BaseTpcpObject, field_type: Union[_ParaTypes, Iterable[_ParaTypes]]
) -> list[str]:
    if isinstance(field_type, _ParaTypes):
        field_type = [field_type]
    return [k for k, v in instance_or_cls.__field_annotations__.items() if v in field_type]


def _has_dangerous_mutable_default(fields: dict[str, inspect.Parameter], cls: type[_BaseTpcpObject]) -> None:
    mutable_defaults = []

    for name, field in fields.items():
        if _is_dangerous_mutable(field):
            # We do not raise an error right here, but wait until we checked everything to provide a more helpful
            # error message.
            mutable_defaults.append(name)

    if len(mutable_defaults) > 0:
        raise MutableDefaultsError(
            f"The class {cls.__name__} contains mutable objects as default values ({mutable_defaults}). "
            "This can lead to unexpected and unpleasant issues! "
            "To solve this the default value should be generated by a factory that produces a new value for each "
            "instance. "
            "If you simply want a new version of the provided default, wrap the value into `tpcp.CloneFactory`"
            "This will enforce a clone/copy of the respective input, whenever a new instance of your Algorithm "
            "or Pipeline is created. "
            "\n"
            "Note, that we do not check for all cases of mutable objects. "
            f"At the moment, we check only for {_get_dangerous_mutable_types()}. "
            "To learn more about this topic, check TODO: LINK."
        )


def _annotations_are_valid(
    field_annotations: dict[str, _ParaTypes],
    fields: dict[str, inspect.Parameter],
    cls_name: str,
) -> None:
    for k, v in field_annotations.items():
        if "__" in k:
            if v is _ParaTypes.SIMPLE:
                warnings.warn(
                    "Annotating a nested parameter (parameter like `nested_object__nest_para` as a simple "
                    "Parameter has no effect and the entire line should be removed.",
                    PotentialUserErrorWarning,
                    stacklevel=2,
                )
        elif k not in fields:
            raise ValueError(
                f"The field '{k}' of {cls_name} was annotated as a `tpcp` "
                "(Hyper/Pure/Normal/Optimizable)-Parameter, but is not a parameter listed in the init! "
                "Add the parameter to the init if it is an actual parameter of your algorithm, or remove the "
                "annotation."
            )


def _has_invalid_name(fields: dict[str, inspect.Parameter], cls: type[_BaseTpcpObject]) -> None:
    invalid_names = [f for f in fields if "__" in f]
    if len(invalid_names) > 0:
        raise ValidationError(
            f"The parameters {invalid_names} of {cls.__name__} have a double-underscore in their name. "
            "This is not allowed, as it interferes with the nested naming conventions in tpcp.\n\n"
            "If you are seeing this while using `dataclasses`, and trying to annotate nested parameters, make sure to "
            "exclude from the init using the explicit `field` syntax or by wrapping it with `ClassVar`:\n\n"
            ">>> @dataclasses.dataclass\n"
            "... class MyClass(BaseTpcpObject):\n"
            "...    # first option\n"
            "...    nested__parameter: OptiPara[int] = dataclasses.field(init=False)\n"
            "...    # second option\n"
            "...    other_nested__parameter: ClassVar[OptiPara[int]]\n\n"
        )
    invalid_names = [f for f in fields if f.endswith("_")]
    if len(invalid_names) > 0:
        raise ValidationError(
            f"The parameters {invalid_names} of {cls.__name__} have a trailing underscore in their name. "
            "This is not allowed, as it interferes naming convention of result objects in tpcp. "
            "Only result values are allowed to have a trailing underscore.\n\n"
            "If you are seeing this while using `dataclasses` or `attrs` and trying to annotate types of result "
            "objects, make sure to exclude them from the init using the explicit `field` syntax:\n\n"
            ">>> @dataclasses.dataclass\n"
            "... class MyClass(BaseTpcpObject):\n"
            "...    result_: int = dataclasses.field(init=False)\n"
        )


def _get_dangerous_mutable_types() -> tuple[type, ...]:
    # TODO: Update this list or even make it a white list?
    return _BaseTpcpObject, list, dict, np.ndarray, pd.DataFrame, BaseEstimator


def _is_dangerous_mutable(field: inspect.Parameter) -> bool:
    """Check if a parameter is one of the mutable objects "considered" dangerous."""
    if field.default is inspect.Parameter.empty or isinstance(field.default, BaseFactory):
        return False
    if isinstance(field.default, _get_dangerous_mutable_types()):
        return True
    return False


def _is_builtin_class_instance(obj: Any) -> bool:
    return type(obj).__module__ == "builtins"


def clone(algorithm: T, *, safe: bool = False) -> T:
    """Construct a new algorithm object with the same parameters.

    This is a modified version from sklearn and the original was published under a BSD-3 license and the original file
    can be found here: https://github.com/scikit-learn/scikit-learn/blob/0d378913b/sklearn/base.py#L31

    The method creates a copy of tpcp algorithms and pipelines, without any results attached.
    I.e. it is equivalent to creating a new instance with the same parameter.
    For all objects that are not tpcp objects (or lists of them), a deepcopy is created.

    .. warning :: This function will not clone sklearn models as expected!
                  `sklearn.clone` will remove the trained model from sklearn object by creating a new instance.
                  This clone method, will deepcopy sklearn models.
                  This means fitted models will be copied and are still available afterwards.
                  For more information have a look at the documentation about "inputs and results" in tpcp.
                  (see :ref:`general concepts <general_concepts>` and :ref:`sklearn differences <sklearn_differences>`)

    Parameters
    ----------
    algorithm : {list, tuple, set} of algorithm instance or a single algorithm instance
        The algorithm or group of algorithms to be cloned.
    safe : bool, default=False
        If safe is False, clone will fall back to a deep copy on objects that are not algorithms.

    """
    if algorithm is NOTHING:
        return algorithm
    # Handle named tuple
    if isinstance(algorithm, tuple) and hasattr(algorithm, "_asdict") and hasattr(algorithm, "_fields"):
        return type(algorithm)(*(clone(a, safe=safe) for a in algorithm))
    if isinstance(algorithm, dict):
        return {k: clone(v, safe=safe) for k, v in algorithm.items()}
    if isinstance(algorithm, (list, tuple, set, frozenset)):
        return type(algorithm)([clone(a, safe=safe) for a in algorithm])

    # Compared to sklearn, we check specifically for BaseTpcpObject and not just if `get_params` is defined on the
    # object.
    # Due to the way algorithms/pipelines in tpcp work, they need to inherit from BaseTpcpObject.
    # Therefore, we check explicitly for that, as we do not want to accidentally treat a sklearn algo (or similar) as
    # algorithm
    if not isinstance(algorithm, BaseTpcpObject):
        if not safe:
            # For some reason, some libraries print stuff to stdout when cloning.
            with Path(os.devnull).open("w") as devnull, contextlib.redirect_stdout(devnull):
                return copy.deepcopy(algorithm)
        raise TypeError(
            f"Cannot clone object '{algorithm!r}' (type {type(algorithm)}): "
            "it does not seem to be a compatible algorithm/pipline class or general `tpcp` object as it does not "
            "inherit from `BaseTpcpObject` or `Algorithm` or `Pipeline`."
        )

    klass = algorithm.__class__
    new_object_params = algorithm.get_params(deep=False)
    for name, param in new_object_params.items():
        # We defer the actual cloning of the parameters to the class itself to allow to handle stupid edge cases on
        # user side without modifying tpcp itself.
        new_object_params[name] = klass.__clone_param__(name, param)
    new_object = klass(**new_object_params)
    params_set = new_object.get_params(deep=False)

    # quick sanity check of the parameters of the clone
    for name in new_object_params:
        param1 = new_object_params[name]
        param2 = params_set[name]
        if param1 is not param2:
            raise RuntimeError(
                f"Cannot clone object {algorithm}, as the constructor either does not set or modifies parameter {name}"
            )
    return new_object


class BaseFactory:
    """Baseclass for factories to circumvent mutable defaults."""

    def get_value(self) -> Any:
        """Provide the value generated by the factory.

        This method will be called every time a new instance of a class is created.
        """
        raise NotImplementedError()

    def __call__(self) -> Any:
        return self.get_value()


class CloneFactory(BaseFactory, Generic[T]):
    """Init factory that creates a clone of the provided default values on instance initialisation.

    This can be used to make sure that each instance get their own version own copy of a mutable default of a class
    init.
    Under the hood this uses :func:`~tpcp.clone`.
    """

    def __init__(self, default_value: T) -> None:
        self.default_value = default_value

    def get_value(self) -> T:
        """Clone the default value for each instance."""
        return clone(self.default_value)

    def __repr__(self) -> str:
        """Print the representation for the factory."""
        return f"cf({self.default_value!r})"


# The typing here is obviously wrong and a hack to make object that are wrapped by clone factory seem to be just of
# the same type.
# Ideally clone factory would be some generic type of proxy, but it isn't.
# Therefore, we fake the return type, so that external type checking in user code works without any errors.
# As long as `cf` is used for its intended purpose that should not matter much.
def cf(default_value: T) -> T:  # pylint: disable=invalid-name
    """Wrap mutable default value with the `CloneFactory`.

    This is basically an alias for :class:`~tpcp.CloneFactory`
    """
    return CloneFactory(default_value)  # type: ignore  # noqa: PGH003


class _Default:
    """Wrapper for default values.

    This can be used to check if an optional function argument was explicitly provided by the user or not.
    See :func:`~tpcp.validate.validate` for  a usage example.
    """

    def __init__(self, value):
        self.value = value

    def get_value(self):
        return self.value

    def __repr__(self):
        """Print the representation of the value."""
        return repr(self.value)
